''' Fuzzy String Matching.

This module is a pretty verbatim Python port of fzf's FuzzyMatchV2
trimmed down to a basic usecase of matching ASCII strings case sensitively.

For more information check out the source, I have not bothered to copy
the introductory comment/documentation:

    https://github.com/junegunn/fzf/blob/b1a0ab8086/src/algo/algo.go

'''
import collections
from dataclasses import dataclass
from enum import Enum
from visidata import VisiData, vd

# Overwrite to true to get some diagnostic visualization
DEBUG = False

scoreMatch        = 16
scoreGapStart     = -3
scoreGapExtension = -1

# We prefer matches at the beginning of a word, but the bonus should not be
# too great to prevent the longer acronym matches from always winning over
# shorter fuzzy matches. The bonus point here was specifically chosen that
# the bonus is cancelled when the gap between the acronyms grows over
# 8 characters, which is approximately the average length of the words found
# in web2 dictionary and my file system.
bonusBoundary = scoreMatch / 2

# Although bonus point for non-word characters is non-contextual, we need it
# for computing bonus points for consecutive chunks starting with a non-word
# character.
bonusNonWord = scoreMatch / 2

# Edge-triggered bonus for matches in camelCase words.
# Compared to word-boundary case, they don't accompany single-character gaps
# (e.g. FooBar vs. foo-bar), so we deduct bonus point accordingly.
bonusCamel123 = bonusBoundary + scoreGapExtension

# Minimum bonus point given to characters in consecutive chunks.
# Note that bonus points for consecutive matches shouldn't have needed if we
# used fixed match score as in the original algorithm.
bonusConsecutive = -(scoreGapStart + scoreGapExtension)

# The first character in the typed pattern usually has more significance
# than the rest so it's important that it appears at special positions where
# bonus points are given, e.g. 'to-go' vs. 'ongoing' on 'og' or on 'ogo'.
# The amount of the extra bonus should be limited so that the gap penalty is
# still respected.
bonusFirstCharMultiplier = 2

# Extra bonus for word boundary after whitespace character or beginning of the string
bonusBoundaryWhite = bonusBoundary + 2

# Extra bonus for word boundary after slash, colon, semi-colon, and comma
bonusBoundaryDelimiter = bonusBoundary + 1

delimiterChars = '/,:;|'

vd.theme_option('color_match', 'red', 'color for matching chars in palette chooser')

whiteChars = ' \t\n\v\f\r\x85\xA0'

(
    charWhite,
    charNonWord,
    charDelimiter,
    charLower,
    charUpper,
    charLetter,
    charNumber,
) = range(7)
initialCharClass = charWhite


def asciiFuzzyIndex(target, pattern):
    '''Return a fuzzy* starting position of the pattern,
    or -1, if pattern isn't a fuzzy match.

    *the position is adapted one back, if possible,
    for bonus determination reasons.
    '''
    first_idx, idx = 0, 0
    for pidx in range(len(pattern)):
        idx = target.find(pattern[pidx], idx)
        if idx < 0:
            return -1
        if pidx == 0 and idx > 0:
            # Step back to find the right bonus point
            first_idx = idx - 1
        idx += 1
    return first_idx


def charClassOfAscii(char):
    if char >= 'a' and char <= 'z':
        return charLower
    elif char >= 'A' and char <= 'Z':
        return charUpper
    elif char >= '0' and char <= '9':
        return charNumber
    elif char in whiteChars:
        return charWhite
    elif char in delimiterChars:
        return charDelimiter
    return charNonWord


def bonusFor(prevClass, class_):
    if class_ > charNonWord:
        if prevClass == charWhite:
            # Word boundary after whitespace
            return bonusBoundaryWhite
        elif prevClass == charDelimiter:
            # Word boundary after a delimiter character
            return bonusBoundaryDelimiter
        elif prevClass == charNonWord:
            # Word boundary
            return bonusBoundary
    if (
        prevClass == charLower
        and class_ == charUpper
        or prevClass != charNumber
        and class_ == charNumber
    ):
        # camelCase letter123
        return bonusCamel123
    elif class_ == charNonWord:
        return bonusNonWord
    elif class_ == charWhite:
        return bonusBoundaryWhite
    return 0


def debugV2(T, pattern, F, lastIdx, H, C):
    '''Visualize the score matrix and matching positions.'''
    width = lastIdx - F[0] + 1

    for i, f in enumerate(F):
        I = i * width
        if i == 0:
            print('  ', end='')
            for j in range(f, lastIdx + 1):
                print(f' {T[j]} ', end='')
            print()
        print(pattern[i] + ' ', end='')
        for idx in range(F[0], f):
            print(' 0 ', end='')
        for idx in range(f, lastIdx + 1):
            print(f'{int(H[i*width+idx-int(F[0])]):2d} ', end='')
        print()

        print('  ', end='')
        for idx, p in enumerate(C[I : I + width]):
            if idx + int(F[0]) < int(F[i]):
                p = 0
            if p > 0:
                print(f'{p:2d} ', end='')
            else:
                print('   ', end='')
        print()


@dataclass
class MatchResult:
    '''Represents a scored match of a fuzzymatching search.

    start: starting index of where the pattern is in the target sequence
    end: Similarly, the end index (exclusive)
    score: A value of how good the match is.
    positions: A list of indices, indexing into the target sequence.
               Corresponds to every position a letter of the pattern was found
               for this particular alignment.
    '''

    start: int
    end: int
    score: int
    positions: 'list[int]'


def _fuzzymatch(target: str, pattern: str) -> MatchResult:
    '''Fuzzy string matching algorithm.

    For a target sequence, check whether (and how good) a pattern is matching.

    Returns a MatchResult, which contains start and end index of the match,
    a score, and the positions where the pattern occurred.

    The matching is case sensitive, so it's necessary to lower input and pattern
    in the caller, if preferred otherwise.

    The functionality is based on fzf's FuzzyMatchV2, minus some advanced features.
    '''
    patternLength = len(pattern)
    if patternLength == 0:
        return MatchResult(0, 0, 0, [])
    targetLength = len(target)

    # Phase 1: Optimized search for ASCII string
    idx = asciiFuzzyIndex(target, pattern)
    if idx < 0:
        return MatchResult(-1, -1, 0, None)

    H0 = [0] * targetLength
    C0 = [0] * targetLength
    # Bonus point for each position
    B = [0] * targetLength
    # The first occurrence of each character in the pattern
    F = [0] * patternLength
    T = list(target)

    # Phase 2: Calculate bonus for each point
    maxScore, maxScorePos = 0, 0
    pidx, lastIdx = 0, 0
    pchar0, pchar, prevH0, prevClass, inGap = (
        pattern[0],
        pattern[0],
        0,
        initialCharClass,
        False,
    )
    Tsub = T[idx:]
    H0sub, C0sub, Bsub = H0[idx:], C0[idx:], B[idx:]

    for off, char in enumerate(Tsub):
        class_ = charClassOfAscii(char)
        bonus = bonusFor(prevClass, class_)
        Bsub[off] = bonus
        prevClass = class_

        if char == pchar:
            if pidx < patternLength:
                F[pidx] = idx + off
                pidx += 1
                pchar = pattern[min(pidx, patternLength - 1)]
            lastIdx = idx + off

        if char == pchar0:
            score = scoreMatch + bonus * bonusFirstCharMultiplier
            H0sub[off] = score
            C0sub[off] = 1
            if patternLength == 1 and (score > maxScore):
                maxScore, maxScorePos = score, idx + off
                if bonus >= bonusBoundary:
                    break
            inGap = False
        else:
            if inGap:
                H0sub[off] = max(prevH0 + scoreGapExtension, 0)
            else:
                H0sub[off] = max(prevH0 + scoreGapStart, 0)
            C0sub[off] = 0
            inGap = True
        prevH0 = H0sub[off]

    # write back, because slices in python are a full copy (as opposed to go)
    H0[idx:], C0[idx:], B[idx:] = H0sub, C0sub, Bsub

    if pidx != patternLength:
        return MatchResult(-1, -1, 0, None)
    if patternLength == 1:
        return MatchResult(maxScorePos, maxScorePos + 1, maxScore, [maxScorePos])

    # Phase 3: Fill in score matrix (H)
    # do not allow omission.
    f0 = F[0]
    width = lastIdx - f0 + 1
    H = [0] * width * patternLength
    H[:width] = list(H0[f0 : lastIdx + 1])

    # Possible length of consecutive chunk at each position.
    C = [0] * width * patternLength
    C[:width] = C0[f0 : lastIdx + 1]

    Fsub = F[1:]
    Psub = pattern[1:]
    for off, f in enumerate(Fsub):
        pchar = Psub[off]
        pidx = off + 1
        row = pidx * width
        inGap = False
        Tsub = T[f : lastIdx + 1]
        Bsub = B[f:][: len(Tsub)]
        H[row + f - f0 - 1] = 0
        for off, char in enumerate(Tsub):
            Cdiag = C[row + f - f0 - 1 - width :][: len(Tsub)]
            Hleft = H[row + f - f0 - 1 :][: len(Tsub)]
            Hdiag = H[row + f - f0 - 1 - width :][: len(Tsub)]
            col = off + f
            s1, s2, consecutive = 0, 0, 0

            if inGap:
                s2 = Hleft[off] + scoreGapExtension
            else:
                s2 = Hleft[off] + scoreGapStart

            if pchar == char:
                s1 = Hdiag[off] + scoreMatch
                b = Bsub[off]
                consecutive = Cdiag[off] + 1
                if consecutive > 1:
                    fb = B[col - consecutive + 1]
                    # Break consecutive chunk
                    if b >= bonusBoundary and b > fb:
                        consecutive = 1
                    else:
                        b = max(b, max(bonusConsecutive, fb))
                if s1 + b < s2:
                    s1 += Bsub[off]
                    consecutive = 0
                else:
                    s1 += b
            C[row + f - f0 + off] = consecutive

            inGap = s1 < s2
            score = max(max(s1, s2), 0)
            if pidx == patternLength - 1 and score > maxScore:
                maxScore, maxScorePos = score, col
            H[row + f - f0 + off] = score

    if DEBUG:
        debugV2(T, pattern, F, lastIdx, H, C)

    # Phase 4. (Optional) Backtrace to find character positions
    pos = []
    i = patternLength - 1
    j = maxScorePos
    preferMatch = True
    while True:
        I = i * width
        j0 = j - f0
        s = H[I + j0]

        s1, s2 = 0, 0
        if i > 0 and j >= int(F[i]):
            s1 = H[I - width + j0 - 1]
        if j > int(F[i]):
            s2 = H[I + j0 - 1]

        if s > s1 and (s > s2 or s == s2 and preferMatch):
            pos.append(j)
            if i == 0:
                break
            i -= 1
        preferMatch = (
            C[I + j0] > 1 or I + width + j0 + 1 < len(C) and C[I + width + j0 + 1] > 0
        )
        j -= 1

    # Start offset we return here is only relevant when begin tiebreak is used.
    # However finding the accurate offset requires backtracking, and we don't
    # want to pay extra cost for the option that has lost its importance.
    return MatchResult(j, maxScorePos + 1, int(maxScore), pos)


def _format_match(s, positions):
    out = list(s)
    for p in positions:
        out[p] = f'[:match]{out[p]}[/]'
    return "".join(out)

CombinedMatch = collections.namedtuple('CombinedMatch', 'score formatted match')


@VisiData.api
def fuzzymatch(vd, haystack:"list[dict[str, str]]", needles:"list[str]) -> list[CombinedMatch]", case_sensitive=False):
    '''Perform matching that is case-insensitive by default. Return sorted list of matching dict values in haystack, augmenting the input dicts with _score:int and _positions:dict[k,set[int]] where k is each non-_ key in the haystack dict. Set *case_sensitive* to match case.'''
    if not case_sensitive:
        needles = [ p.lower() for p in needles]
    matches = []
    for h in haystack:
        match = {}
        formatted_hay = {}
        for k, v in h.items():
            if k[0] == '_': continue
            positions = set()
            v_match = v if case_sensitive else v.lower()
            for p in needles:
                mr = _fuzzymatch(v_match, p)
                if mr.score > 0:
                    match.setdefault(k, []).append(mr)
                    positions |= set(mr.positions)
            formatted_hay[k] = _format_match(v, positions)

        if match:
            # square to prefer larger scores in a single haystack
            score = int(sum([mr.score**2 for mrs in match.values() for mr in mrs]))
            matches.append(CombinedMatch(score=score, formatted=formatted_hay, match=h))

    return sorted(matches, key=lambda m: -m.score)


@VisiData.api
def test_fuzzymatch(vd):
    assert asciiFuzzyIndex('helo', 'h') == 0
    assert asciiFuzzyIndex('helo', 'hlo') == 0
    assert asciiFuzzyIndex('helo', 'e') == 0
    assert asciiFuzzyIndex('helo', 'el') == 0
    assert asciiFuzzyIndex('helo', 'eo') == 0
    assert asciiFuzzyIndex('helo', 'l') == 1
    assert asciiFuzzyIndex('helo', 'lo') == 1
    assert asciiFuzzyIndex('helo', 'o') == 2
    assert asciiFuzzyIndex('helo', 'ooh') == -1

    assert charClassOfAscii('a') == charLower
    assert charClassOfAscii('C') == charUpper
    assert charClassOfAscii('2') == charNumber
    assert charClassOfAscii(' ') == charWhite
    assert charClassOfAscii(',') == charDelimiter

    assert _fuzzymatch('hello', '') == MatchResult(0, 0, 0, [])
    assert _fuzzymatch('hello', 'nono') == MatchResult(-1, -1, 0, None)
    assert _fuzzymatch('hello', 'l') == MatchResult(2, 3, 16, [2])
    assert _fuzzymatch('hello world', 'elo wo') == MatchResult(
        1, 8, 127, [7, 6, 5, 4, 2, 1]
    )



# Copyright (c) 2022 Saul Pwanson
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""
usage: unzip_http [-h] [-l] [-f] [-o] url [files ...]

Extract individual files from .zip files over http without downloading the
entire archive. HTTP server must send `Accept-Ranges: bytes` and
`Content-Length` in headers.

positional arguments:
  url                   URL of the remote zip file
  files                 Files to extract. If no filenames given, displays .zip
                        contents (filenames and sizes). Each filename can be a
                        wildcard glob.

options:
  -h, --help            show this help message and exit
  -l, --list            List files in the remote zip file
  -f, --full-filepaths  Recreate folder structure from zip file when extracting
                        (instead of extracting the files to the current
                        directory)
  -o, --stdout          Write files to stdout (if multiple files: concatenate
                        them to stdout, in zipfile order)
"""

import sys
import os
import io
import math
import time
import zlib
import struct
import fnmatch
import argparse
import pathlib
import urllib.parse
from visidata import vd


__version__ = '0.6'


def error(s):
    raise Exception(s)

def warning(s):
    print(s, file=sys.stderr)

def get_bits(val:int, *args):
    'Generate bitfields (one for each arg) from LSB to MSB.'
    for n in args:
        x = val & (2**n-1)
        val >>= n
        yield x


class RemoteZipInfo:
    def __init__(self, filename:str='',
                       date_time:int = 0,
                       header_offset:int = 0,
                       compress_type:int = 0,
                       compress_size:int = 0,
                       file_size:int = 0):
        self.filename = filename
        self.header_offset = header_offset
        self.compress_type = compress_type
        self.compress_size = compress_size
        self.file_size = file_size

        sec, mins, hour, day, mon, year = get_bits(date_time, 5, 6, 5, 5, 4, 7)
        self.date_time = (year+1980, mon, day, hour, mins, sec)

    def is_dir(self):
        return self.filename.endswith('/')

    def parse_extra(self, extra):
        i = 0
        while i < len(extra):
            fieldid, fieldsz = struct.unpack_from('<HH', extra, i)
            i += 4

            if fieldid == 0x0001:  # ZIP64
                if fieldsz == 8: fmt = '<Q'
                elif fieldsz == 16: fmt = '<QQ'
                elif fieldsz == 24: fmt = '<QQQ'
                elif fieldsz == 28: fmt = '<QQQI'

                vals = list(struct.unpack_from(fmt, extra, i))
                if self.file_size == 0xffffffff:
                    self.file_size = vals.pop(0)

                if self.compress_size == 0xffffffff:
                    self.compress_size = vals.pop(0)

                if self.header_offset == 0xffffffff:
                    self.header_offset = vals.pop(0)

            i += fieldsz


class RemoteZipFile:
    fmt_eocd = '<IHHHHIIH'  # end of central directory
    fmt_eocd64 = '<IQHHIIQQQQ'  # end of central directory ZIP64
    fmt_cdirentry = '<IHHHHIIIIHHHHHII'  # central directory entry
    fmt_localhdr = '<IHHHIIIIHH'  # local directory header
    magic_eocd64 = b'\x50\x4b\x06\x06'
    magic_eocd = b'\x50\x4b\x05\x06'

    def __init__(self, url):
        urllib3 = vd.importExternal('urllib3')
        self.url = url
        self.http = urllib3.PoolManager()
        self.zip_size = 0

    def __enter__(self):
        return self

    def __exit__(self, a, b, c):
        pass

    @property
    def files(self):
        if not hasattr(self, '_files'):
            self._files = {r.filename:r for r in self.infoiter()}
        return self._files

    def infolist(self):
        return list(self.infoiter())

    def namelist(self):
        return list(r.filename for r in self.infoiter())

    def infoiter(self):
        resp = self.http.request('HEAD', self.url)
        r = resp.headers.get('Accept-Ranges', '')
        if r != 'bytes':
            hostname = urllib.parse.urlparse(self.url).netloc
            warning(f"{hostname} Accept-Ranges header ('{r}') is not 'bytes'--trying anyway")

        self.zip_size = int(resp.headers['Content-Length'])
        resp = self.get_range(
            max(self.zip_size-65536, 0),
            65536
        )

        cdir_start = -1
        i = resp.data.rfind(self.magic_eocd64)
        if i >= 0:
            magic, eocd_sz, create_ver, min_ver, disk_num, disk_start, disk_num_records, total_num_records, \
                cdir_bytes, cdir_start = struct.unpack_from(self.fmt_eocd64, resp.data, offset=i)
        else:
            i = resp.data.rfind(self.magic_eocd)
            if i >= 0:
                magic, \
                    disk_num, disk_start, disk_num_records, total_num_records, \
                    cdir_bytes, cdir_start, comment_len = struct.unpack_from(self.fmt_eocd, resp.data, offset=i)

        if cdir_start < 0 or cdir_start >= self.zip_size:
            error('cannot find central directory')

        if self.zip_size <= 65536:
            filehdr_index = cdir_start
        else:
            filehdr_index = 65536 - (self.zip_size - cdir_start)

        if filehdr_index < 0:
            resp = self.get_range(cdir_start, self.zip_size - cdir_start)
            filehdr_index = 0

        cdir_end = filehdr_index + cdir_bytes
        while filehdr_index < cdir_end:
            sizeof_cdirentry = struct.calcsize(self.fmt_cdirentry)

            magic, ver, ver_needed, flags, method, date_time, crc, \
                complen, uncomplen, fnlen, extralen, commentlen, \
                disknum_start, internal_attr, external_attr, local_header_ofs = \
                    struct.unpack_from(self.fmt_cdirentry, resp.data, offset=filehdr_index)

            filehdr_index += sizeof_cdirentry

            filename = resp.data[filehdr_index:filehdr_index+fnlen]
            filehdr_index += fnlen

            extra = resp.data[filehdr_index:filehdr_index+extralen]
            filehdr_index += extralen

            comment = resp.data[filehdr_index:filehdr_index+commentlen]
            filehdr_index += commentlen

            rzi = RemoteZipInfo(filename.decode(), date_time, local_header_ofs, method, complen, uncomplen)

            rzi.parse_extra(extra)
            yield rzi

    def extract(self, member, path=None, pwd=None):
            if pwd:
                raise NotImplementedError('Passwords not supported yet')

            path = path or pathlib.Path('.')

            outpath = path/member
            os.makedirs(outpath.parent, exist_ok=True)
            with self._open(member) as fpin:
                with open(path/member, mode='wb') as fpout:
                    while True:
                        r = fpin.read(65536)
                        if not r:
                            break
                        fpout.write(r)

    def extractall(self, path=None, members=None, pwd=None):
        for fn in members or self.namelist():
            self.extract(fn, path, pwd=pwd)

    def get_range(self, start, n):
        return self.http.request('GET', self.url, headers={'Range': f'bytes={start}-{start+n-1}'}, preload_content=False)

    def matching_files(self, *globs):
        for f in self.files.values():
            if any(fnmatch.fnmatch(f.filename, g) for g in globs):
                yield f

    def _open(self, fn):
        if isinstance(fn, str):
            f = list(self.matching_files(fn))
            if not f:
                error(f'no files matching {fn}')
            f = f[0]
        else:
            f = fn

        sizeof_localhdr = struct.calcsize(self.fmt_localhdr)
        r = self.get_range(f.header_offset, sizeof_localhdr)
        localhdr = struct.unpack_from(self.fmt_localhdr, r.data)
        magic, ver, flags, method, dos_datetime, _, _, uncomplen, fnlen, extralen = localhdr
        if method == 0: # none
            return self.get_range(f.header_offset + sizeof_localhdr + fnlen + extralen, f.compress_size)
        elif method == 8: # DEFLATE
            resp = self.get_range(f.header_offset + sizeof_localhdr + fnlen + extralen, f.compress_size)
            return io.BufferedReader(RemoteZipStream(resp, f))
        else:
            error(f'unknown compression method {method}')

    def open(self, fn):
        return io.TextIOWrapper(self._open(fn))


class RemoteZipStream(io.RawIOBase):
    def __init__(self, fp, info):
        super().__init__()
        self.raw = fp
        self._decompressor = zlib.decompressobj(-15)
        self._buffer = bytes()

    def readable(self):
        return True

    def readinto(self, b):
        r = self.read(len(b))
        b[:len(r)] = r
        return len(r)

    def read(self, n):
        while n > len(self._buffer):
            r = self.raw.read(2**18)
            if not r:
                self._buffer += self._decompressor.flush()
                break
            self._buffer += self._decompressor.decompress(r)

        ret = self._buffer[:n]
        self._buffer = self._buffer[n:]

        return ret


 ### script start

class StreamProgress:
    def __init__(self, fp, name='', total=0):
        self.name = name
        self.fp = fp
        self.total = total
        self.start_time = time.time()
        self.last_update = 0
        self.amtread = 0

    def read(self, n):
        r = self.fp.read(n)
        self.amtread += len(r)
        now = time.time()
        if now - self.last_update > 0.1:
            self.last_update = now

            elapsed_s = now - self.start_time
            sys.stderr.write(f'\r{elapsed_s:.0f}s  {self.amtread/10**6:.02f}/{self.total/10**6:.02f}MB  ({self.amtread/10**6/elapsed_s:.02f} MB/s)  {self.name}')

        if not r:
            sys.stderr.write('\n')

        return r


def list_files(rzf):
    def safelog(x):
        return 1 if x == 0 else math.ceil(math.log10(x))

    digits_compr = max(safelog(f.compress_size) for f in rzf.infolist())
    digits_plain = max(safelog(f.file_size    ) for f in rzf.infolist())
    fmtstr = f'%{digits_compr}d -> %{digits_plain}d\t%s'
    for f in rzf.infolist():
        print(fmtstr % (f.compress_size, f.file_size, f.filename), file=sys.stderr)


def extract_one(outfile, rzf, f, ofname):
    print(f'Extracting {f.filename} to {ofname}...', file=sys.stderr)

    fp = StreamProgress(rzf.open(f), name=f.filename, total=f.compress_size)
    while r := fp.read(2**18):
        outfile.write(r)


def download_file(f, rzf, args):
    if not any(fnmatch.fnmatch(f.filename, g) for g in args.files):
        return

    if args.stdout:
        extract_one(sys.stdout.buffer, rzf, f, "stdout")
    else:
        path = pathlib.Path(f.filename)
        if args.full_filepaths:
            path.parent.mkdir(parents=True, exist_ok=True)
        else:
            path = path.name

        with open(str(path), 'wb') as of:
            extract_one(of, rzf, f, str(path))


def main():
    parser = argparse.ArgumentParser(prog='unzip-http', \
        description="Extract individual files from .zip files over http without downloading the entire archive. HTTP server must send `Accept-Ranges: bytes` and `Content-Length` in headers.")

    parser.add_argument('-l', '--list', action='store_true', default=False,
                        help="List files in the remote zip file")
    parser.add_argument('-f', '--full-filepaths', action='store_true', default=False,
                        help="Recreate folder structure from zip file when extracting (instead of extracting the files to the current directory)")
    parser.add_argument('-o', '--stdout', action='store_true', default=False,
                        help="Write files to stdout (if multiple files: concatenate them to stdout, in zipfile order)")

    parser.add_argument("url", nargs=1, help="URL of the remote zip file")
    parser.add_argument("files", nargs='*', help="Files to extract. If no filenames given, displays .zip contents (filenames and sizes). Each filename can be a wildcard glob.")

    args = parser.parse_args()

    rzf = RemoteZipFile(args.url[0])
    if args.list or len(args.files) == 0:
        list_files(rzf)
    else:
        for f in rzf.infolist():
            download_file(f, rzf, args)



if __name__ == '__main__':
    main()



from copy import copy
from visidata import vd, asyncthread, Progress, Sheet, Column, options, UNLOADED, ColumnsSheet
import re

cmdlog_col_prefix='\u241f'  #string ␟ to mark the start of column info in an ordering string

@Sheet.api
def orderBy(sheet, *cols, reverse=False, change_column=False, save_cmd_input=True):
    '''Add *cols* to internal ordering and re-sort the rows accordingly.
    Pass *reverse* as True to order these *cols* descending.
    Pass empty *cols* (or cols[0] of None) to clear internal ordering.
    Set *change_column* to True to change the sort status of a single column: add/remove/invert it.
    When changing a column, *cols* must have length 1. Sort columns that had higher priority are unchanged. Lower-priority columns are removed.
    If *change_column* is False, *cols* will be added to the existing ordering.
    If *save_cmd_input* is True, the full ordering that results will be saved in the cmdlog for future replay in the 'input' parameter.
    '''

    if options.undo:
        vd.addUndo(setattr, sheet, '_ordering', copy(sheet._ordering))
        vd.addUndo(setattr, sheet, 'rows', copy(sheet.rows))

    # for replay, read the full column ordering from the cmdlog input parameter  #2688
    input = vd.getLastArgs()
    if input:
        sheet._ordering = order_from_string(sheet, input)
        sheet.sort()
        if save_cmd_input:
            vd.activeCommand.input = order_string(sheet)
        return

    do_sort = False
    if not cols or cols[0] is None:
        sheet._ordering.clear()
        cols = cols[1:]
        do_sort = True

    if change_column:
        if len(cols) > 1:
            vd.fail('sort order edit must only be applied to a single column')
        new_ordering = edit_ordering(sheet._ordering, cols[0], reverse)
        sheet._ordering = new_ordering
        do_sort = True
    else:
        for c in cols:
            sheet._ordering.append((c, reverse))
            do_sort = True

    if do_sort:
        sheet.sort()
    if save_cmd_input:
        vd.activeCommand.input = order_string(sheet)

class Reversor:
    def __init__(self, obj):
        self.obj = obj

    def __eq__(self, other):
        return other.obj == self.obj

    def __lt__(self, other):
        return other.obj < self.obj

def order_string(sheet):
    sheet._ordering = sheet.ordering  #converts any ambiguous colname strings to unambiguous Column objects
    ret = ''.join([cmdlog_col_prefix+('>' if reverse else '<') + str(col.name) for col, reverse in sheet._ordering])
    return ret

def order_from_string(sheet, s):
    instructions = re.split(cmdlog_col_prefix + '(?=[<>])', s)[1:]
    ordering = []
    for instr in instructions:
        c = sheet.column(instr[1:])
        if instr[0] == '<':
            reverse = False
        elif instr[0] == '>':
            reverse = True
        ordering.append((c, reverse))
    return ordering

def edit_ordering(ordering, col, reverse):
    '''Return a modified ordering based on editing a single column *col*:   add it, remove it, or flip its direction.
    Columns after *col* in the ordering (with lower sort priority) are also removed from the ordering.
    *ordering* is a list of tuples:  (Column, boolean), where the boolean defines the sort direction.
    '''
    new_ordering = []
    # handle changes to status of columns that are already in the ordering:  add/remove/flip
    changed = False
    for c, old_reverse in ordering:
        if c is col:
            if reverse != old_reverse: # reverse the column's sort direction
                new_ordering.append((c, reverse))
            # if the sort direction is unchanged, remove the column from the ordering
            changed = True
            # columns after the edited column will be dropped from the ordering
            break
        new_ordering.append((c, old_reverse))
    if not changed:
        new_ordering.append((col, reverse))
    return new_ordering

@Sheet.cached_property
def ordering(sheet) -> 'list[tuple[Column, bool]]':
    ret = []
    for col, reverse in sheet._ordering:
        if isinstance(col, str):
            col = sheet.column(col)
        ret.append((col, reverse))
    return ret


@Sheet.api
def sortkey(sheet, r, ordering:'list[tuple[Column, bool]]'=[]):
    ret = []
    for col, reverse in (ordering or sheet.ordering):
        val = col.getTypedValue(r)
        ret.append(Reversor(val) if reverse else val)

    return ret


@Sheet.api
@asyncthread
def sort(self):
    'Sort rows according to the current internal ordering.'
    if self.rows is UNLOADED:
        return
    try:
        with Progress(gerund='sorting', total=self.nRows) as prog:
            # replace ambiguous colname strings with unambiguous Column objects  #2494
            self._ordering = self.ordering
            def _sortkey(r):
                prog.addProgress(1)
                return self.sortkey(r, ordering=self._ordering)

            # must not reassign self.rows: use .sort() instead of sorted()
            self.rows.sort(key=_sortkey)
    except TypeError as e:
        vd.warning('sort incomplete due to TypeError; change column type')
        vd.exceptionCaught(e, status=False)

ColumnsSheet.columns += [
        Column('sortorder',
            type=int,
            getter=lambda c,r: _sort_order(c, r),
            help='sort priority and direction in source sheet')
]

def _sort_order(col, srccol):
    sort_cols = [(n+1, reverse) for n, (c, reverse) in enumerate(srccol.sheet.ordering) if c is srccol]
    if not sort_cols:
        return None
    n, reverse = sort_cols[0]
    return -n if reverse else n


# replace existing sort criteria
Sheet.addCommand('[', 'sort-asc', 'orderBy(None, cursorCol)', 'sort ascending by current column; replace any existing sort criteria')
Sheet.addCommand(']', 'sort-desc', 'orderBy(None, cursorCol, reverse=True)', 'sort descending by current column; replace any existing sort criteria ')
Sheet.addCommand('g[', 'sort-keys-asc', 'orderBy(None, *keyCols)', 'sort ascending by all key columns; replace any existing sort criteria')
Sheet.addCommand('g]', 'sort-keys-desc', 'orderBy(None, *keyCols, reverse=True)', 'sort descending by all key columns; replace any existing sort criteria')

# add to existing sort criteria
Sheet.addCommand('', 'sort-asc-add', 'orderBy(cursorCol)', 'sort ascending by current column; add to existing sort criteria')
Sheet.addCommand('', 'sort-desc-add', 'orderBy(cursorCol, reverse=True)', 'sort descending by current column; add to existing sort criteria')
Sheet.addCommand('z[', 'sort-asc-change', 'orderBy(cursorCol, change_column=True)', 'sort ascending by current column; keep higher priority sort criteria')
Sheet.addCommand('z]', 'sort-desc-change', 'orderBy(cursorCol, reverse=True, change_column=True)', 'sort descending by current column; keep higher priority sort criteria')
Sheet.addCommand('gz[', 'sort-keys-asc-add', 'orderBy(*keyCols)', 'sort ascending by all key columns; add to existing sort criteria')
Sheet.addCommand('gz]', 'sort-keys-desc-add', 'orderBy(*keyCols, reverse=True)', 'sort descending by all key columns; add to existing sort criteria')

vd.addMenuItems('''
    Column > Sort by > current column only > ascending > sort-asc
    Column > Sort by > current column only > descending > sort-desc
    Column > Sort by > current column also > ascending > sort-asc-add
    Column > Sort by > current column also > descending > sort-desc-add
    Column > Sort by > key columns > ascending > sort-keys-asc
    Column > Sort by > key columns > descending > sort-keys-desc
''')



import json
import time
import collections
import random
from functools import cached_property
from base64 import b64encode

from visidata import vd, VisiData, Canvas, Animation, Path, asyncthread, clipdraw, colors, ItemColumn, Sheet, wraptext

vd.theme_option('color_hint', 'black on yellow', '')

@VisiData.api
def getNoahsPath(vd, name):
    return Path(vd.pkg_resources_files('visidata')/f'experimental/noahs_tapestry/{name}')

@VisiData.api
def openNoahsText(vd, name):
    return vd.getNoahsPath(name).open(encoding='utf-8')

@VisiData.cached_property
def noahsDatabase(vd):
    return vd.open_sqlite(vd.getNoahsPath('noahs.sqlite'))

class NoahsPuzzle(Sheet):
    guide = '''
        # Puzzle {sheet.puznum}
        - `Shift+A` to input a solution to the puzzle
        - `Shift+Y` to attempt the current cell as the solution
        - `Shift+B` to open Noah's Database Backup
        - `Shift+V` to view Noah's Tapestry
    '''
    rowtype = 'lines'  # rowdef: [linenum, text]
    filetype = 'txt'
    columns = [
        ItemColumn('linenum', 0, type=int, width=0),
        ItemColumn('text', 1, width=80, displayer='full'),
    ]
    precious = False

    def iterload(self):
        clues = json.loads(vd.getNoahsPath(f'clues.json').read_text())
        source = vd.getNoahsPath(f'puzzle{self.puznum}.md')
        winWidth = 78
        formatted_text = source.open(encoding='utf-8').read().format(**clues)
        for startingLine, text in enumerate(formatted_text.splitlines()):
            text = text.strip()
            if text:
                for i, (L, _) in enumerate(wraptext(str(text), width=winWidth)):
                    yield [startingLine+i+1, L]
            else:
                yield [startingLine+1, text]

@VisiData.cached_property
def noahsSolutions(vd):
    return json.loads(vd.getNoahsPath(f'solutions.json').read_text())

@VisiData.api
def solve_puzzle(vd, answer):
    puznum = vd.noahsCurrentPuznum
    if b64encode(answer.encode()).decode() != vd.noahsSolutions[f'p{puznum}']:
        vd.fail("Hmmm, that doesn't seem right. Try again?")

    vd.noahsTapestry.solved.add(puznum)
    vd.status(f'Correct! The candle is now lit.')
    vd.push(vd.noahsTapestry)


class Tapestry(Canvas):
    @property
    def guide(self):
        ret = ''
        if vd.screenWidth < 120 or vd.screenHeight < 50:
            ret = f'''
            # [:black on yellow]WARNING: TERMINAL TOO SMALL[/]
            Please expand your terminal to at least 120x50 (currently {vd.screenWidth}x{vd.screenHeight})
        '''

        ret += '''
            # Noah's Tapestry
            An interactive data game

            - `Shift+N` to open the next puzzle
        '''

        return ret

    solved = set()
    def reload(self):
        self.noahs_menorah = Animation(vd.openNoahsText('menorah.ddw'))
        self.noahs_tapestry = Animation(vd.openNoahsText('tapestry.ddw'))
        self.noahs_flame = Animation(vd.openNoahsText('flame.ddw'))
        self.solved = set()

#        self.keep_running()

    @asyncthread
    def keep_running(self):
        while True:
            time.sleep(1)

    def draw(self, scr):
        solvedays = ['menorah']+[f'day{i}' for i in self.solved]
        t = time.time()
        self.noahs_menorah.draw(scr, t=0, y=30, x=19)
        if not self.solved:
            clipdraw(scr, 22, 52, "Light the [:italic]shamash[/]", colors['255'])
            return

        self.noahs_menorah.draw(scr, t=0.1, y=30, x=19, tags=solvedays)
        self.noahs_tapestry.draw(scr, t=t, tags=solvedays)

        for i in self.solved:
            xs = [58, 22+9*8, 22+9*7, 22+9*6, 22+9*5, 22+27, 22+18, 22+9, 22, 22]
            ys = [28, 32, 32, 32, 32, 32, 32, 32, 32, 32]
            x = xs[i]
            y = ys[i]

            self.noahs_flame.draw(scr, t=t+(i+random.random())*0.2, y=y, x=x)

    def open_puzzle(self, puznum=None):
        if puznum is None:
            puznum = 0
            if self.solved:
                puznum = max(self.solved)+1

        vs = NoahsPuzzle('puzzle', str(puznum), source=self, puznum=puznum)
        vs.ensureLoaded()
        vd.noahsCurrentPuznum = puznum
        return vs


@VisiData.lazy_property
def noahsTapestry(vd):
    vd.curses_timeout = 50
    return Tapestry('noahs', 'tapestry')


NoahsPuzzle.options.color_default = '178 yellow on 232 black'

vd.addCommand('Shift+B', 'open-noahs-database', 'vd.push(noahsDatabase)', "open database for Noah's Tapestry")
vd.addCommand('Shift+V', 'open-noahs-tapestry', 'vd.push(noahsTapestry)', "open Noah's Tapestry")
Tapestry.addCommand('Shift+N', 'open-puzzle-next', 'vd.push(open_puzzle())', 'open next unsolved puzzle')
NoahsPuzzle.addCommand('Shift+A', 'solve-puzzle-input', 'solve_puzzle(input("Answer: "))', 'input an answer to the current puzzle')
Sheet.addCommand('Shift+Y', 'solve-puzzle-cell', 'solve_puzzle(cursorValue)', 'input an answer to the current puzzle')
for i in range(9):
    Tapestry.addCommand(f'{i}', f'open-puzzle-{i}', f'vd.push(open_puzzle({i}))')
    Tapestry.addCommand(f'Alt+{i}', f'solve-puzzle-force-{i}', f'sheet.solved.add({i})')



from visidata import vd, Menu
from .gitsheet import GitSheet


@GitSheet.api
def abortWhatever(sheet):
    inp = sheet.gitInProgress()
    if inp.startswith('cherry-pick'):
        sheet.modifyGit('cherry-pick', '--abort')
    elif inp.startswith('merg'):
        sheet.modifyGit('merge', '--abort')
    elif inp.startswith('bisect'):
        sheet.modifyGit('bisect', 'reset')
    elif inp.startswith('rebas') or inp.startswith('apply'):
        sheet.modifyGit('rebase', '--abort')  # or --quit?
    else:
        vd.status('nothing to abort')


GitSheet.addCommand('^A', 'git-abort', 'abortWhatever()', 'abort the current in-progress action')


vd.addMenuItems('Git > Abort > git-abort')



from visidata import VisiData, Sheet, vd


vd.option('incr_base', 1.0, 'start value for column increments', replay=True)


@VisiData.api
def numrange(vd, n, step=1):
    'Generate n values, starting from options.incr_base and increasing by step for each number.'
    base = type(step)(vd.options.incr_base)
    yield from ((base+x)*step for x in range(n))


def test_numrange(vd=None):
    assert list(vd.numrange(5)) == [1,2,3,4,5]
    assert list(vd.numrange(5, step=5)) == [5,10,15,20,25]  #2769

@VisiData.api
def num(vd, *args):
    'Return parsed string as number, preferring int to float.'
    try:
        return int(*args)
    except Exception:
        return float(*args)


Sheet.addCommand('i', 'addcol-incr', 'c=SettableColumn(type=int); addColumnAtCursor(c); c.setValues(rows, *numrange(nRows))', 'add column with incremental values')
Sheet.addCommand('gi', 'setcol-incr', 'cursorCol.setValues(selectedRows, *numrange(sheet.nSelectedRows))', 'set current column for selected rows to incremental values')
Sheet.addCommand('zi', 'addcol-incr-step', 'n=num(input("interval step: ")); c=SettableColumn(type=type(n)); addColumnAtCursor(c); c.setValues(rows, *numrange(nRows, step=n))', 'add column with incremental values times given step')
Sheet.addCommand('gzi', 'setcol-incr-step', 'n=num(input("interval step: ")); cursorCol.setValues(selectedRows, *numrange(nSelectedRows, n))', 'set current column for selected rows to incremental values times given step')

vd.addMenuItems('''
    Column > Add column > increment > addcol-incr
    Edit > Modify > selected cells > increment > setcol-incr
''')



import json
from collections import Counter

from visidata import vd, date, anytype, VisiData, PyobjSheet, AttrDict, stacktrace, TypedExceptionWrapper, AlwaysDict, ItemColumn, wrapply, TypedWrapper, Progress, Sheet

vd.option('json_indent', None, 'indent to use when saving json')
vd.option('json_sort_keys', False, 'sort object keys when saving to json')
vd.option('json_ensure_ascii', True, 'ensure ascii encode when saving json')
vd.option('default_colname', '', 'column name to use for non-dict rows')

@VisiData.api
def guess_json(vd, p):
    with p.open(encoding=vd.options.encoding) as fp:
        try:
            line = next(fp)
        except StopIteration:
            return

    line = line.strip()

    if line.startswith('{') and line.endswith('}'):
        return dict(filetype='jsonl')

    if line.startswith(tuple('[{')):
        return dict(filetype='json')


@VisiData.api
def open_jsonobj(vd, p):
    return JsonSheet(p.base_stem, source=p)

@VisiData.api
def open_jsonl(vd, p):
    return JsonSheet(p.base_stem, source=p)

VisiData.open_ndjson = VisiData.open_ldjson = VisiData.open_json = VisiData.open_jsonl


class JsonSheet(Sheet):
    _rowtype = AttrDict
    def resetCols(self):
        self._knownKeys = set()
        super().resetCols()

    def iterload(self):
        with self.open_text_source() as fp:
            for L in fp:
                L = L.strip()
                try:
                    if not L: # skip blank lines
                        continue
                    ret = json.loads(L, object_hook=AttrDict)
                    if isinstance(ret, list):
                        yield from ret
                    else:
                        yield ret

                except ValueError as e:
                    if self.rows:   # if any rows have been added already
                        e.stacktrace = stacktrace()
                        yield TypedExceptionWrapper(json.loads, L, exception=e)  # an error on one line
                    else:
                        with self.open_text_source() as fp:
                            ret = json.load(fp, object_hook=AttrDict)
                            if isinstance(ret, list):
                                yield from ret
                            else:
                                yield ret
                        break

    def addColumn(self, *cols, index=None):
        for c in cols:
            self._knownKeys.add(c.expr or c.name)
        return super().addColumn(*cols, index=index)

    def addRow(self, row, index=None):
        # Wrap non-dict rows in a dummy object with a predictable key name.
        # This allows for more consistent handling of rows containing scalars
        # or lists.
        if not isinstance(row, dict):
            v = {self.options.default_colname: row}
            row = AlwaysDict(row, **v)

        ret = super().addRow(row, index=index)

        for k in row:
            if k not in self._knownKeys:
                c = ItemColumn(k)
                self.addColumn(c)

        return ret

    def newRow(self, **fields):
        return AttrDict(fields)

    def openRow(self, row):
        return PyobjSheet("%s[%s]" % (self.name, self.rowname(row)), source=row)

JsonSheet.init('_knownKeys', set, copy=True)  # set of row keys already seen

## saving json and jsonl

class _vjsonEncoder(json.JSONEncoder):
    def default(self, obj):
        return str(obj)


@VisiData.api
def get_json_value(vd, col, row):
    o = wrapply(col.getTypedValue, row)
    if isinstance(o, TypedExceptionWrapper):
        o = col.sheet.options.safe_error or str(o.exception)
    elif isinstance(o, TypedWrapper):
        o = o.val
    elif isinstance(o, date):
        o = col.getDisplayValue(row)
    return o


def _rowdict(cols, row, keep_nulls=False):
    ret = {}
    for col in cols:
        o = vd.get_json_value(col, row)
        if keep_nulls or o is not None:
            ret[col.name] = o
    return ret


@VisiData.api
def encode_json(vd, row, cols, enc=_vjsonEncoder(sort_keys=False)):
    'Return JSON string for given *row* and given *cols*.'
    return enc.encode(_rowdict(cols, row))


@VisiData.api
def save_json(vd, p, *vsheets):
    vs = vsheets[0]
    with p.open(mode='w', encoding=vs.options.save_encoding) as fp:
        try:
            indent = int(vs.options.json_indent)
        except Exception:
            indent = vs.options.json_indent

        jsonenc = _vjsonEncoder(indent=indent, sort_keys=vs.options.json_sort_keys, ensure_ascii=vs.options.json_ensure_ascii)

        dupnames = find_duplicates([vs.name for vs in vsheets])
        for name in dupnames:
            vd.warning('json cannot save sheet with duplicated name: ' + name)
        for vs in vsheets:
            dupnames = find_duplicates([c.name for c in vs.visibleCols])
            for name in dupnames:
                vd.warning('json cannot save column with duplicated name: ' + name)
        if len(vsheets) == 1:
            fp.write('[\n')
            vs = vsheets[0]
            with Progress(gerund='saving'):
                for i, row in enumerate(vs.iterrows()):
                    if i > 0:
                        fp.write(',\n')
                    rd = _rowdict(vs.visibleCols, row, keep_nulls=(i==0))
                    fp.write(jsonenc.encode(rd))
            fp.write('\n]\n')
        else:
            it = {vs.name: [_rowdict(vs.visibleCols, row, keep_nulls=(i==0)) for i, row in enumerate(vs.iterrows())] for vs in vsheets}

            with Progress(gerund='saving'):
                for chunk in jsonenc.iterencode(it):
                    fp.write(chunk)


@Sheet.api
def write_jsonl(vs, fp):
        vcols = vs.visibleCols
        jsonenc = _vjsonEncoder()
        dupnames = find_duplicates([c.name for c in vcols])
        for name in dupnames:
            vd.warning('json cannot save column with duplicated name: ' + name)
        with Progress(gerund='saving'):
            for i, row in enumerate(vs.iterrows()):
                rowdict = _rowdict(vcols, row, keep_nulls=(i==0))
                fp.write(jsonenc.encode(rowdict) + '\n')

        if len(vs) == 0:
            vd.warning(
                "Output file is empty - cannot save headers without data for jsonl.\n"
                "Use `.jsonla` filetype to save as JSONL arrays format "
                "rather than JSONL dict format to preserve the headers."
            )


@VisiData.api
def save_jsonl(vd, p, *vsheets):
    with p.open(mode='w', encoding=vsheets[0].options.save_encoding) as fp:
        if len(vsheets) > 1:
            vd.warning('jsonl cannot separate sheets yet. Concatenating all rows.')
        for vs in vsheets:
            vs.write_jsonl(fp)


@VisiData.api
def JSON(vd, s:str):
    'Parse `s` as JSON.'
    return json.loads(s)

def find_duplicates(names):
    return list(colname for colname,count in Counter(names).items() if count > 1)

JsonSheet.options.encoding = 'utf-8'
JsonSheet.options.regex_skip = r'^(//|#).*'

VisiData.save_ndjson = VisiData.save_jsonl
VisiData.save_ldjson = VisiData.save_jsonl

vd.addGlobals({
    'JsonSheet': JsonSheet,
    'JsonLinesSheet': JsonSheet,
})





from visidata import vd, VisiData, Sheet, IndexSheet, SequenceSheet, ColumnItem, Path, AttrDict, ColumnAttr, asyncthread


def _google_creds_fn():

    filename = 'google_creds.json'
    google_creds_path = vd.pkg_resources_files('vdplus.api.google') / filename
    import os
    if not os.path.exists(google_creds_path):
        vd.error(f'{filename} file does not exist in {google_creds_path.parent}\n'
                 'Create it by following this guide: https://github.com/saulpw/visidata/blob/develop/docs/gmail.md')
    else:
        return str(google_creds_path)


@VisiData.api
def google_auth(vd, scopes=None):
    import pickle
    import os.path
    import urllib.parse

    SCOPES = []
    for scope in scopes.split():
        if not scope.startswith('https://'):
            scope = 'https://www.googleapis.com/auth/' + scope
        SCOPES.append(scope)

    GOOGLE_TOKEN_FILE = Path(vd.options.visidata_dir)/f'google-{urllib.parse.quote_plus(str(scopes))}.pickle'
    creds = None
    if os.path.exists(GOOGLE_TOKEN_FILE):
        with open(GOOGLE_TOKEN_FILE, 'rb') as fp:
            creds = pickle.load(fp)

    if not creds or not creds.valid:
        if creds and creds.expired and creds.refresh_token:
            from google.auth.transport.requests import Request
            creds.refresh(Request())
        else:
            from google_auth_oauthlib.flow import InstalledAppFlow
            flow = InstalledAppFlow.from_client_secrets_file(_google_creds_fn(), SCOPES)
            creds = flow.run_local_server(port=0)

        with open(GOOGLE_TOKEN_FILE, 'wb') as fp:
            pickle.dump(creds, fp)

    return creds



from itertools import islice

from visidata import vd, Column, VisiData, ItemColumn, AttrColumn, Path, AttrDict

from .gitsheet import GitSheet


CONFIG_CONTEXTS = ('local', 'global', 'system')

@VisiData.api
def git_config(vd, p, args):
    if not args or '-l' in args:
        return GitConfig('git-config', source=p)

vd.git_options = vd.git_config

def batched(iterable, n=1):
    "Batch data into tuples of length n. The last batch may be shorter."
    # batched('ABCDEFG', 3) --> ABC DEF G
    assert n >= 1, 'n must be at least one'

    while (batch := tuple(islice(iter(iterable), n))):
        yield batch


class GitConfigColumn(Column):
    def calcValue(self, row):
        return row.get(self.expr)

    def putValue(self, row, val):
        if val is None:
            self.sheet.loggit('config', '--unset', '--'+self.expr, row['option'])
        else:
            self.sheet.loggit('config', '--'+self.expr, row['option'], val)

        row[self.expr] = val


class GitConfig(GitSheet):
    guide = '''
        # git config
        Add, edit, or delete git config options.

        - Make changes using standard commands like `a`dd and `e`dit
        - `z Ctrl+S` to commit changes to git config file.
    '''
    rowtype = 'git options'  # rowdef: [scope, origin, opt, val]
    defer = True
    columns = [
        ItemColumn('option', width=20),
    ]
    nKeys = 1

    def iterload(self):
        cmd = self.git_iter('config', '--list', '--show-scope', '--show-origin', '-z')
        self.gitopts = {}
        scopes = {c.name:c for c in self.columns}
        for row in batched(cmd, 3):
            if len(row) < 3:
                break

            scope, origin, optval = row
            opt, val = optval.split('\n', 1)

            if opt in self.gitopts:
                r = self.gitopts[opt]
            else:
                r = AttrDict(option=opt)
                self.gitopts[opt] = r
                yield r

            r[scope] = val

            if scope not in scopes:
                c = GitConfigColumn(scope, expr=scope)
                self.addColumn(c)
                scopes[scope] = c

    def commitDeleteRow(self, row):
        for k in CONFIG_CONTEXTS:
            if row.get(k):
                self.loggit('config', '--unset', '--'+k, row['option'])

    def commitAddRow(self, row):
        for k in CONFIG_CONTEXTS:
            if row.get(k):
                self.loggit('config', '--add', '--'+k, row['option'], row.get(k))


GitSheet.addCommand('gO', 'git-config', 'vd.push(GitConfig("git_config", source=Path(".")))', 'push sheet of git options')


vd.addMenuItems('''
    Git > Config > git-config
''')



'''motd: display a low-priority random Message Of The Day on startup.

Call `domotd()` to spawn an asyncthread to read and/or fetch
a motd file from a url.  The file may be text or unheaded TSV, with one message per row in the first column.

Any Exception ends the thread silently.

options.motd_url may be set to another URL, or empty string to disable entirely.
'''

import random

from visidata import options, asyncsingle, vd, VisiData

vd.option('motd_url', 'https://visidata.org/motd-'+vd.version, 'source of randomized startup messages', sheettype=None)


vd.motd = 'Support VisiData: https://github.com/sponsors/saulpw'

@VisiData.api
@asyncsingle
def domotd(vd):
    try:
        if options.motd_url:
            p = vd.urlcache(options.motd_url, days=1)
            line = random.choice(list(p))
            vd.motd = line.split('\t')[0]
    except Exception:
        pass
    vd.status(vd.motd, priority=-1)



from visidata import VisiData, vd, Sheet, Column, Progress, ColumnAttr, ColumnItem, SubColumnItem, InvertedCanvas


@VisiData.api
def open_ttf(vd, p):
    return TTFTablesSheet(p.base_stem, source=p)

vd.open_otf = vd.open_ttf

class TTFTablesSheet(Sheet):
    rowtype = 'font tables'
    columns = [
        ColumnAttr('cmap'),
        ColumnAttr('format', type=int),
        ColumnAttr('language', type=int),
        ColumnAttr('length', type=int),
        ColumnAttr('platEncID', type=int),
        ColumnAttr('platformID', type=int),
        Column('isSymbol', getter=lambda col,row: row.isSymbol()),
        Column('isUnicode', getter=lambda col,row: row.isUnicode()),
    ]

    def openRow(self, row):
        return TTFGlyphsSheet(self.name+'_glyphs', source=self, sourceRows=[row], ttf=self.ttf)

    def iterload(self):
        fontTools = vd.importExternal('fontTools.ttLib', 'fonttools')

        self.ttf = fontTools.TTFont(str(self.source), 0, allowVID=0, ignoreDecompileErrors=True, fontNumber=-1)
        for cmap in self.ttf["cmap"].tables:
            yield cmap


class TTFGlyphsSheet(Sheet):
    rowtype = 'glyphs'  # rowdef: (codepoint, glyphid, fontTools.ttLib.ttFont._TTGlyphGlyf)
    columns = [
        ColumnItem('codepoint', 0, type=int, fmtstr='%0X'),
        ColumnItem('glyphid', 1),
        SubColumnItem(2, ColumnAttr('height', type=int)),
        SubColumnItem(2, ColumnAttr('width', type=int)),
        SubColumnItem(2, ColumnAttr('lsb')),
        SubColumnItem(2, ColumnAttr('tsb')),
    ]

    def openRow(self, row):
        return makePen(self.name+"_"+row[1], source=row[2], glyphSet=self.ttf.getGlyphSet())

    def iterload(self):
        glyphs = self.ttf.getGlyphSet()
        for cmap in self.sourceRows:
            for codepoint, glyphid in Progress(cmap.cmap.items(), total=len(cmap.cmap)):
                yield (codepoint, glyphid, glyphs[glyphid])


def makePen(*args, **kwargs):
    fontTools = vd.importExternal('fontTools', 'fonttools')
    from fontTools.pens.basePen import BasePen

    class GlyphPen(InvertedCanvas, BasePen):
        aspectRatio = 1.0
        def __init__(self, name, **kwargs):
            super().__init__(name, **kwargs)
            self.path_firstxy = None
            self.lastxy = None
            self.attr = self.plotColor(('glyph',))

        def _moveTo(self, xy):
            self.lastxy = xy
            if self.path_firstxy is None:
                self.path_firstxy = xy

        def _lineTo(self, xy):
            x1, y1 = self.lastxy
            x2, y2 = xy
            self.line(x1, y1, x2, y2, self.attr)
            self._moveTo(xy)

        def _closePath(self):
            if self.path_firstxy:
                if (self.path_firstxy != self.lastxy):
                    self._lineTo(self.path_firstxy)
                self.path_firstxy = None
            self.lastxy = None

        def _endPath(self):
            self.path_firstxy = None
            self.lastxy = None

        def _curveToOne(self, xy1, xy2, xy3):
            vd.error('NotImplemented')

        def _qCurveToOne(self, xy1, xy2):
            self.qcurve([self.lastxy, xy1, xy2], self.attr)
            self._moveTo(xy2)

        def reload(self):
            self.reset()
            self.source.draw(self)
            self.refresh()

    return GlyphPen(*args, **kwargs)


#TTFGlyphsSheet.bindkey('.', 'open-row')



import string
import visidata


visidata.vd.prettykeys_trdict = {
        ' ': 'Space',  # must be first
        '^[': 'Alt+',
        '^J': 'Enter',
        '^M': 'Enter',
        '^I': 'Tab',
        'KEY_BTAB': 'Shift+Tab',
        '^@': 'Ctrl+Space',
        'KEY_UP':    'Up',
        'KEY_DOWN':  'Down',
        'KEY_LEFT':  'Left',
        'KEY_RIGHT': 'Right',
        'KEY_HOME':  'Home',
        'KEY_END':   'End',
        'KEY_EOL':   'End',
        'KEY_PPAGE': 'PgUp',
        'KEY_NPAGE': 'PgDn',

        'kUP':       'Shift+Up',
        'kDN':       'Shift+Down',
        'kUP5':      'Ctrl+Up',
        'kDN5':      'Ctrl+Down',
        'kLFT5':     'Ctrl+Left',
        'kRIT5':     'Ctrl+Right',
        'kHOM5':     'Ctrl+Home',
        'kEND5':     'Ctrl+End',
        'kPRV5':     'Ctrl+PgUp',
        'kNXT5':     'Ctrl+PgDn',
        'KEY_IC5':   'Ctrl+Ins',
        'KEY_DC5':   'Ctrl+Del',
        'kDC5':      'Ctrl+Del',
        'KEY_SDC':   'Shift+Del',

        'KEY_IC':    'Ins',
        'KEY_DC':    'Del',

        'KEY_SRIGHT':'Shift+Right',
        'KEY_SR':    'Shift+Up',
        'KEY_SF':    'Shift+Down',
        'KEY_SLEFT': 'Shift+Left',
        'KEY_SHOME': 'Shift+Home',
        'KEY_SEND':  'Shift+End',
        'KEY_SPREVIOUS': 'Shift+PgUp',
        'KEY_SNEXT': 'Shift+PgDn',

        'kxIN': 'FocusIn',
        'kxOUT': 'FocusOut',

        'KEY_BACKSPACE': 'Bksp',
        'BUTTON1_RELEASED': 'LeftBtnUp',
        'BUTTON2_RELEASED': 'MiddleBtnUp',
        'BUTTON3_RELEASED': 'RightBtnUp',
        'BUTTON1_PRESSED': 'LeftClick',
        'BUTTON2_PRESSED': 'MiddleClick',
        'BUTTON3_PRESSED': 'RightClick',
        'BUTTON4_PRESSED': 'ScrollUp',
        'BUTTON5_PRESSED': 'ScrollDown',
        'REPORT_MOUSE_POSITION': 'ScrollDown',
        '2097152': 'ScrollDown',
    }

for i in range(1, 13):
    d = visidata.vd.prettykeys_trdict
    d[f'KEY_F({i})'] = f'F{i}'
    d[f'KEY_F({i+12})'] = f'Shift+F{i}'
    d[f'KEY_F({i+24})'] = f'Ctrl+F{i}'
    d[f'KEY_F({i+36})'] = f'Ctrl+Shift+F{i}'
    d[f'KEY_F({i+48})'] = f'Alt+F{i}'
    d[f'KEY_F({i+60})'] = f'Alt+Shift+F{i}'


@visidata.VisiData.api
def prettykeys(vd, key):
    if not key or '+' in key[:-1]:
        return key

    for k, v in vd.prettykeys_trdict.items():
        key = key.replace(k, v)

    # replace ^ with Ctrl but not if ^ is last char
    key = key[:-1].replace('^', 'Ctrl+')+key[-1]
    # 1497: allow Shift+ for Alt keys
    if key[-1] in string.ascii_uppercase and ('+' not in key or 'Alt+' in key) and '_' not in key:
        key = key[:-1] + 'Shift+' + key[-1]

    return key.strip()



'''Provide USD(s) function to convert string like '£300' or '205 AUD' to equivalent US$ as float.
Uses data from api.apilayer.com/fixer. Requires an API key for apilayer.com.
'''

from visidata import vd
import functools
import json

vd.option('fixer_api_key', '', 'API Key for api.apilayer.com/fixer')
vd.option('fixer_cache_days', 1, 'Cache days for currency conversions')

currency_symbols = {
    '$': 'USD',
    '£': 'GBP',
    '₩': 'KRW',
    '€': 'EUR',
    '₪': 'ILS',
    'zł': 'PLN',
    '₽': 'RUB',
    '₫': 'VND',
}

def currency_rates_json(date='latest', base='USD'):
    url = 'https://api.apilayer.com/fixer/%s?base=%s' % (date, base)
    return vd.urlcache(
        url,
        days=vd.options.fixer_cache_days,
        headers={
            # First need to set some additional headers as otherwise apilayers will block it with a 403
            # See also https://stackoverflow.com/questions/13303449/urllib2-httperror-http-error-403-forbidden
            'User-Agent': 'Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.11 (KHTML, like Gecko) Chrome/23.0.1271.64 Safari/537.11',
            'Accept': 'text/html,application/xhtml+xml,application/xml;q=0.9,*/*;q=0.8',
            'Accept-Charset': 'ISO-8859-1,utf-8;q=0.7,*;q=0.3',
            'Accept-Encoding': 'none',
            'Accept-Language': 'en-US,en;q=0.8',
            'Connection': 'keep-alive',

            # Finally set Apikey
            'apikey': vd.options.fixer_api_key
        }
    ).read_text()

@functools.lru_cache()
def currency_rates():
    return json.loads(currency_rates_json())['rates']

@functools.lru_cache()
def currency_multiplier(src_currency, dest_currency):
    'returns equivalent value in USD for an amt of currency_code'
    if src_currency == 'USD':
        return 1.0
    eur_usd_mult = currency_rates()['USD']
    eur_src_mult = currency_rates()[src_currency]
    usd_mult = eur_usd_mult/eur_src_mult
    if dest_currency == 'USD':
        return usd_mult
    
    return usd_mult/currency_rates()[dest_currency]

def USD(s):
    for currency_symbol, currency_code in currency_symbols.items():
        if currency_symbol in s:
            amt = float(s.replace(currency_symbol, ''))
            return amt*currency_multiplier(currency_code, 'USD')

    amtstr, currcode = s.split(' ')
    return float(amtstr) * currency_multiplier(currcode, 'USD')


vd.addGlobals(USD=USD)



import traceback
import sys
import re

from visidata import vd, VisiData

vd.option('debug', False, 'exit on error and display stacktrace')


class ExpectedException(Exception):
    'Controlled Exception from fail() or confirm().  Status or other interface update is done by raiser.'
    pass


def stacktrace(e=None, exclude_caller=False):
    '''Return a list of strings for the stack trace, without newlines
    at the end. If an exception handler is executing, and *e* is none,
    the stack trace includes extra levels of callers beyond the level
    where the exception was caught. If *exclude_caller* is True, the
    trace will exclude the function that called stacktrace(). The
    trace will exclude several uninformative levels that are run
    in interactive visidata.'''

    if e:
        return traceback.format_exception_only(type(e), e)
    #in Python 3.11 we can replace sys.exc_info() with sys.exception()
    handling = (sys.exc_info() != (None, None, None))

    stack = ''.join(traceback.format_stack()).strip().splitlines()

    if handling:
        trim_levels = 2       # remove levels for stacktrace() -> format_stack()
        if exclude_caller:
            trim_levels += 1
        trace_above = stack[:-2*trim_levels]
    else:
        trace_above = stack
    if trace_above:
        trace_above[0] = '  ' + trace_above[0]  #fix indent level of first line
    try:
        # remove several levels of uninformative stacktrace in typical interactive vd
        idx = trace_above.index('    ret = vd.mainloop(scr)')
        trace_above = trace_above[idx+1:]
    except ValueError:
        pass
    if not handling:
        return trace_above
    # remove lines that mark error columns with carets and sometimes tildes
    trace_below = [ line for line in traceback.format_exc().strip().splitlines() if not re.match('^ *~*\\^+$', line) ]
    # move the "Traceback (most recent call last) header to the top of the output
    return [trace_below[0]] + trace_above + trace_below[1:]


@VisiData.api
def exceptionCaught(vd, exc=None, status=True, **kwargs):
    'Add *exc* to list of last errors and add to status history.  Show on left status bar if *status* is True.  Reraise exception if options.debug is True.'
    if isinstance(exc, ExpectedException):  # already reported, don't log
        return
    # save a stack trace that does not include this function
    vd.lastErrors.append(stacktrace(exclude_caller=True))
    if status:
        vd.status(f'{type(exc).__name__}: {exc}', priority=2)
    else:
        vd.addToStatusHistory(vd.lastErrors[-1][-1])
    if vd.options.debug:
        raise


vd.addGlobals(stacktrace=stacktrace, ExpectedException=ExpectedException)

# see textsheet.py for ErrorSheet and associated commands



from visidata import VisiData, vd, BaseSheet

vd._parentscrs = {}  # scr -> parentscr


@VisiData.api
def subwindow(vd, scr, x, y, w, h):
    'Return subwindow with its (0,0) at (x,y) relative to parent scr.  Replacement for scr.derwin() to track parent scr.'
    newscr = scr.derwin(h, w, y, x)
    vd._parentscrs[newscr] = scr
    return newscr


@VisiData.api
def getrootxy(vd, scr):  # like scr.getparyx() but for all ancestor scrs
    px, py = 0, 0
    while scr in vd._parentscrs:
        dy, dx = scr.getparyx()
        if dy > 0: py += dy
        if dx > 0: px += dx
        scr = vd._parentscrs[scr]
    return px, py


vd.bindkey('Alt+[Shift+I', 'no-op')  #2247 focus-in
vd.bindkey('Alt+[Shift+O', 'no-op')  # focus-out



from copy import copy
from visidata import vd, asyncthread, Path, Sheet, IndexSheet, TableSheet


@Sheet.api
def syseditCells(sheet, cols, rows, filetype=None):
    filetype = filetype or vd.input("edit %d %s as filetype: " % (len(rows), sheet.rowtype), value=sheet.options.save_filetype or 'tsv')
    vd.sync(sheet.syseditCells_async(cols, rows, filetype))


@Sheet.api
@asyncthread
def syseditCells_async(sheet, cols, rows, filetype=None):
    vs = copy(sheet)
    vs.rows = rows or vd.fail('no %s selected' % sheet.rowtype)
    vs.columns = cols

    import tempfile
    with tempfile.NamedTemporaryFile() as temp:
        temp.close()  #2118
        p = Path(temp.name)

        vd.status(f'copying {vs.nRows} {vs.rowtype} to {p} as {filetype}')
        vd.sync(vd.saveSheets(p, vs))

        tempvs = vd.openSource(p, filetype=filetype)

        vd.launchExternalEditorPath(p)
        tempvs.source = p
        vd.sync(tempvs.ensureLoaded())

        while isinstance(tempvs, IndexSheet):
            vd.sync(tempvs.ensureLoaded())
            tempvs = tempvs.rows[0]

        for col in sheet.visibleCols:
            tempcol = tempvs.colsByName.get(col.name)
            if not tempcol: # column not in edited version
                continue
            # only assign values that were changed by the editor
            edited_rows = []
            edited_vals = []
            for r, r_edited in zip(rows, tempvs.rows):
                v = tempcol.getDisplayValue(r_edited)
                if col.getDisplayValue(r) != v:
                    edited_rows.append(r)
                    edited_vals.append(v)
            if edited_rows:
                col.setValuesTyped(edited_rows, *edited_vals)


TableSheet.addCommand('^O', 'sysedit-cell', 'cd = cursorDisplay; e = vd.launchExternalEditor(cd); cursorCol.setValues([cursorRow], e) if e != cd else None', 'edit current cell in external $EDITOR')
Sheet.addCommand('g^O', 'sysedit-selected', 'syseditCells(visibleCols, onlySelectedRows)', 'edit rows in $EDITOR')



from visidata import vd, Sheet, options, Column, asyncthread, Progress, PivotGroupRow, ENTER, HistogramColumn

from visidata.loaders._pandas import PandasSheet
from visidata.pivot import PivotSheet

class DataFrameRowSliceAdapter:
    """Tracks original dataframe and a boolean row mask

    This is a workaround to (1) save memory (2) keep id(row)
    consistent when iterating, as id() is used significantly
    by visidata's selectRow implementation.
    """
    def __init__(self, df, mask):
        pd = vd.importExternal('pandas')
        np = vd.importExternal('numpy')
        if not isinstance(df, pd.DataFrame):
            vd.fail('%s is not a dataframe' % type(df).__name__)
        if not isinstance(mask, pd.Series):
            vd.fail('mask %s is not a Series' % type(mask).__name__)
        if df.shape[0] != mask.shape[0]:
            vd.fail('dataframe and mask have different shapes (%s vs %s)' % (df.shape[0], mask.shape[0]))

        self.df = df
        self.mask_bool = mask  # boolean mask
        self.mask_iloc = np.where(mask.values)[0]  # integer indexes corresponding to mask
        self.mask_count = mask.sum()

    def __len__(self):
        return self.mask_count

    def __getitem__(self, k):
        if isinstance(k, slice):
            import pandas as pd
            new_mask = pd.Series(False, index=self.df.index)
            new_mask.iloc[self.mask_iloc[k]] = True
            return DataFrameRowSliceAdapter(self.df, new_mask)
        return self.df.iloc[self.mask_iloc[k]]

    def __iter__(self):
        # With the internal selection API used by PandasSheet,
        # this should no longer be needed and can be replaced by
        # DataFrameAdapter(self.df[self.mask_iloc])
        return DataFrameRowSliceIter(self.df, self.mask_iloc)

    def __getattr__(self, k):
        # This is trouble ..
        return getattr(self.df[self.mask_bool], k)

class DataFrameRowSliceIter:
    def __init__(self, df, mask_iloc, index=0):
        self.df = df
        self.mask_iloc = mask_iloc
        self.index = index

    def __next__(self):
        # Accessing row of original dataframe, to ensure
        # that no copies are made and id() of selected rows
        # will match original dataframe's rows
        if self.index >= self.mask_iloc.shape[0]:
            raise StopIteration()
        row = self.df.iloc[self.mask_iloc[self.index]]
        self.index += 1
        return row

def makePandasFreqTable(sheet, *groupByCols):
    fqcolname = '%s_freq' % '-'.join(col.name for col in groupByCols)
    return PandasFreqTableSheet(sheet.name, fqcolname, groupByCols=groupByCols, source=sheet)


class PandasFreqTableSheet(PivotSheet):
    'Generate frequency-table sheet on currently selected column.'
    rowtype = 'bins'  # rowdef FreqRow(keys, sourcerows)

    def selectRow(self, row):
        # Select all entries in the bin on the source sheet.
        # Use the internally defined _selectByLoc to avoid
        # looping which causes a significant performance hit.
        self.source._selectByILoc(row.sourcerows.mask_iloc, selected=True)
        # then select the bin itself on this sheet
        return super().selectRow(row)

    def unselectRow(self, row):
        self.source._selectByILoc(row.sourcerows.mask_iloc, selected=False)
        return super().unselectRow(row)

    def addUndoSelection(self):
        self.source.addUndoSelection()
        super().addUndoSelection()

    def updateLargest(self, grouprow):
        self.largest = max(self.largest, len(grouprow.sourcerows))

    def loader(self):
        'Generate frequency table then reverse-sort by length.'
        import pandas as pd

        # Note: visidata's base FrequencyTable bins numeric data in ranges
        # (e.g. as a histogram). We currently don't provide support for this
        # for PandasSheet, although we could implement it with a pd.Grouper
        # that operates similarly to pd.cut.
        df = self.source.df.copy()

        # Implementation (special case): for one row, this degenerates
        # to .value_counts(); however this does not order in a stable manner.
        # if len(self.groupByCols) == 1:
        #     this_column = df.loc[:, str(self.groupByCols[0].name)]
        #     value_counts = this_column.value_counts()
        if len(self.groupByCols) >= 1:
            # Implementation (1): add a dummy column to aggregate over in a pd.pivot_table.
            # Is there a way to avoid having to mutate the dataframe? We can delete the
            # column afterwards but we do incur the overhead of block consolidation.
            _pivot_count_column = "__vd_pivot_count"
            if _pivot_count_column not in df.columns:
                df[_pivot_count_column] = 1
            # Aggregate count over columns to group, and then apply a stable sort
            value_counts = df.pivot_table(
                index=[c.name for c in self.groupByCols],
                values=_pivot_count_column,
                aggfunc="count"
            )[_pivot_count_column].sort_values(ascending=False, kind="mergesort")
            # TODO: it seems that the ascending=False causes this to do a "reversed stable sort"?
            # TODO: possibly register something to delete this column as soon as
            # we exit visidata?
            # del df["__vd_pivot_count"]

            # Implementation (2) which does not require adding a dummy column:
            # Compute cross-tabulation to get counts, and sort/remove zero-entries.
            # Note that this is not space-efficient: the initial cross-tabulation will
            # have space on the order of product of number of unique elements for each
            # column, even though its possible the combinations present are sparse
            # and most combinations have zero count.
            # this_column = df.loc[:, str(self.groupByCols[0].name)]
            # value_counts = pd.crosstab(this_column, [df.df[c.name] for c in self.groupByCols[1:]])
            # value_counts = value_counts.stack(list(range(len(self.groupByCols) - 1)))
            # value_counts = value_counts.loc[value_counts > 0].sort_values(ascending=False)
        else:
            vd.fail("Unable to do FrequencyTable, no columns to group on provided")

        # add default bonus columns
        for c in [
                    Column('count', type=int,
                           getter=lambda col,row: len(row.sourcerows)),
                    Column('percent', type=float,
                           getter=lambda col,row: len(row.sourcerows)*100/df.shape[0]),
                    HistogramColumn('histogram', type=str, width=self.options.default_width*2)
                    ]:
            self.addColumn(c)

        for element in Progress(value_counts.index):
            if len(self.groupByCols) == 1:
                element = (element,)
            elif len(element) != len(self.groupByCols):
                vd.fail('different number of index cols and groupby cols (%s vs %s)' % (len(element), len(self.groupByCols)))

            mask = df[self.groupByCols[0].name] == element[0]
            for i in range(1, len(self.groupByCols)):
                mask = mask & (df[self.groupByCols[i].name] == element[i])

            self.addRow(PivotGroupRow(
                element,
                (0, 0),
                DataFrameRowSliceAdapter(df, mask),
                {}
            ))

    def openRow(self, row):
        return self.source.expand_source_rows(row)

@Sheet.api
def expand_source_rows(sheet, row):
    """Support for expanding a row of frequency table to underlying rows"""
    if row.sourcerows is None:
        vd.fail("no source rows")
    return PandasSheet(sheet.name, vd.valueNames(row.discrete_keys, row.numeric_key), source=row.sourcerows)

PandasSheet.addCommand('F', 'freq-col', 'vd.push(makePandasFreqTable(sheet, cursorCol))', 'open Frequency Table grouped on current column, with aggregations of other columns')
PandasSheet.addCommand('gF', 'freq-keys', 'vd.push(makePandasFreqTable(sheet, *keyCols))', 'open Frequency Table grouped by all key columns on source sheet, with aggregations of other columns')

PandasFreqTableSheet.init('largest', lambda: 1)
PandasFreqTableSheet.options.numeric_binning = False

vd.addGlobals(makePandasFreqTable=makePandasFreqTable)



import collections
from copy import copy
from visidata import ScopedSetattr, Column, Sheet, asyncthread, Progress, forward, wrapply, INPROGRESS
from visidata import vlen, vd, date, setitem, anytype
import visidata


# discrete_keys = tuple of formatted discrete keys that group the row
# numeric_key is a range
# sourcerows is list(all source.rows in group)
# pivotrows is { pivot_values: list(source.rows in group with pivot_values) }
PivotGroupRow = collections.namedtuple('PivotGroupRow', 'discrete_keys numeric_key sourcerows pivotrows'.split())

def makePivot(source, groupByCols, pivotCols):
    return PivotSheet('',
            groupByCols=groupByCols,
            pivotCols=pivotCols,
            source=source)

def makeErrorKey(col):
    if col.type is date:
        return date.min # date('2000-01-01')
    else:
        return col.type()

def formatRange(col, numeric_key):
    a, b = numeric_key
    nankey = makeErrorKey(col)
    if b is None:
        return a
    if a is nankey and b is nankey:
        return '#ERR'
    elif a == b:
        return col.format(a)
    return ' - '.join(col.format(x) for x in numeric_key)


class RangeColumn(Column):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.formatter = 'range'

    def formatter_range(self, fmtstr):
        return self._format

    def _format(self, typedval, *args, **kwargs):
        if typedval is None:
            return None
        return formatRange(self.origcol, typedval)


class AggrColumn(Column):
    def calcValue(col, row):
        if col.sheet.loading:
            return visidata.INPROGRESS
        return col.aggregator.aggregate(col.origCol, row.sourcerows)


def makeAggrColumn(aggcol, aggregator):
    aggname = '%s_%s' % (aggcol.name, aggregator.name)

    return AggrColumn(aggname,
                  type=aggregator.type or aggcol.type,
                  fmtstr=aggcol.fmtstr,
                  origCol=aggcol,
                  aggregator=aggregator)


class PivotSheet(Sheet):
    'Summarize key columns in pivot table and display as new sheet.'
    rowtype = 'grouped rows'  # rowdef: PivotGroupRow

    def __init__(self, *names, groupByCols=[], pivotCols=[], **kwargs):
        super().__init__(*names,
                pivotCols=pivotCols, # whose values become columns
                groupByCols=groupByCols,  # whose values become rows
                **kwargs)

    def isNumericRange(self, col):
        return vd.isNumeric(col) and self.source.options.numeric_binning

    def resetCols(self):
        super().resetCols()

        # add key columns (grouped by)
        for colnum, c in enumerate(self.groupByCols):
            if c in self.pivotCols:
                continue

            if self.isNumericRange(c):
                newcol = RangeColumn(c.name, origcol=c, width=c.width and c.width*2, getter=lambda c,r: r.numeric_key)
            else:
                newcol = Column(c.name, width=c.width, fmtstr=c.fmtstr,
                                  type=c.type if c.type in vd.typemap else anytype,
                                  origcol=c,
                                  getter=lambda col,row,i=colnum: row.discrete_keys[i],
                                  setter=lambda col,row,val,i=colnum: setitem(row.discrete_keys, i, val) and col.origcol.setValues(row.sourcerows, val))

            self.addColumn(newcol)

        self.setKeys(self.columns)

    def openRow(self, row):
        'open sheet of source rows aggregated in current pivot row'
        vs = copy(self.source)
        vs.name += "_%s"%"+".join(map(str, row.discrete_keys))
        vs.rows = sum(row.pivotrows.values(), [])
        return vs

    def openCell(self, col, row):
        'open sheet of source rows aggregated in current pivot cell'
        vs = copy(self.source)
        vs.name += "_%s"%col.aggvalue
        vs.rows = row.pivotrows.get(col.aggvalue, [])
        return vs

    def loader(self):
        # two different threads for better interactive display
        vd.sync(self.addAggregateCols(),
                self.groupRows())

    @asyncthread
    def addAggregateCols(self):
        # add aggregated columns
        aggcols = {  # [Column] -> list(aggregators)
            sourcecol: sourcecol.aggregators
                for sourcecol in self.source.visibleCols
                    if sourcecol.aggregators
        } or {  # if pivot given but no aggregators specified
            sourcecol: [vd.aggregators["count"]]
                for sourcecol in self.pivotCols
        }

        if not aggcols:
#            self.addColumn(ColumnAttr('count', 'sourcerows', type=vlen))
            return

        # aggregators without pivot
        if not self.pivotCols:
            for aggcol, aggregatorlist in aggcols.items():
                for aggregator in aggregatorlist:
                    c = makeAggrColumn(aggcol, aggregator)
                    self.addColumn(c)

        # add pivoted columns
        for pivotcol in self.pivotCols:
            allValues = set()
            for value in Progress(pivotcol.getValues(self.source.rows), 'pivoting', total=len(self.source.rows)):
                if value in allValues:
                    continue
                allValues.add(value)

                if len(self.pivotCols) > 1:
                    valname = '%s_%s' % (pivotcol.name, value)
                else:
                    valname = str(value)

                for aggcol, aggregatorlist in aggcols.items():
                    for aggregator in aggregatorlist:
                        if len(aggcols) > 1: #  if more than one aggregated column, include that column name in the new column name
                            aggname = '%s_%s' % (aggcol.name, aggregator.name)
                        else:
                            aggname = aggregator.name


                        if len(aggregatorlist) > 1 or len(aggcols) > 1:
                            colname = '%s_%s' % (aggname, valname)
                            if not self.name:
                                self.name = self.source.name+'_pivot_'+''.join(c.name for c in self.pivotCols)
                        else:
                            colname = valname
                            if not self.name:
                                self.name = self.source.name+'_pivot_'+''.join(c.name for c in self.pivotCols) + '_' + aggname

                        c = Column(colname,
                                    type=aggregator.type or aggcol.type,
                                    aggvalue=value,
                                    getter=lambda col,row,aggcol=aggcol,agg=aggregator: agg.aggregate(aggcol, row.pivotrows.get(col.aggvalue, [])))
                        self.addColumn(c)

#                    if aggregator.name != 'count':  # already have count above
#                        c = Column('Total_' + aggcol.name,
#                                    type=aggregator.type or aggcol.type,
#                                    getter=lambda col,row,aggcol=aggcol,agg=aggregator: agg.aggregate(aggcol, row.sourcerows))
#                        self.addColumn(c)

    @asyncthread
    def groupRows(self, rowfunc=None):
      with ScopedSetattr(self, 'loading', True):
        self.rows = []

        discreteCols = [c for c in self.groupByCols if not self.isNumericRange(c)]

        numericCols = [c for c in self.groupByCols if self.isNumericRange(c)]

        if len(numericCols) > 1:
            vd.fail('only one numeric column can be binned')

        numericBins = []
        degenerateBinning = False
        if numericCols:
            nbins = self.source.options.histogram_bins or int(len(self.source.rows) ** (1./2))
            vals = tuple(numericCols[0].getValues(self.source.rows))
            minval = min(vals) if vals else 0
            maxval = max(vals) if vals else 0
            width = (maxval - minval)/nbins

            if width == 0:
                if vals:
                    # only one value
                    numericBins = [(minval, maxval)]
                else:
                    numericBins = []
            elif (numericCols[0].type in (int, vlen) and nbins > (maxval - minval)) or (width == 1):
                # (more bins than int vals) or (if bins are of width 1), just use the vals as bins
                degenerateBinning = True
                numericBins = [(val, val) for val in sorted(set(vals))]
                nbins = len(numericBins)
            else:
                numericBins = [(minval+width*i, minval+width*(i+1)) for i in range(nbins)]

        # group rows by their keys (groupByCols), and separate by their pivot values (pivotCols)
        groups = {}  # [formattedDiscreteKeys] -> (numericGroupRows:dict(formattedNumericKeyRange -> PivotGroupRow), groupRow:PivotGroupRow)  # groupRow is main/error row

        for sourcerow in Progress(self.source.iterrows(), 'grouping', total=self.source.nRows):
            discreteKeys = list(forward(origcol.getTypedValue(sourcerow)) for origcol in discreteCols)

            # wrapply will pass-through a key-able TypedWrapper
            formattedDiscreteKeys = tuple(wrapply(c.format, v) for v, c in zip(discreteKeys, discreteCols))

            numericGroupRows, groupRow = groups.get(formattedDiscreteKeys, (None, None))
            if numericGroupRows is None:
                # add new group rows
                numericGroupRows = {formatRange(numericCols[0], numRange): PivotGroupRow(discreteKeys, numRange, [], {}) for numRange in numericBins}
                groups[formattedDiscreteKeys] = (numericGroupRows, None)
                for r in numericGroupRows.values():
                    self.addRow(r)

            # find the grouprow this sourcerow belongs in, by numericbin
            if numericCols:
                try:
                    val = numericCols[0].getValue(sourcerow)
                    val = wrapply(numericCols[0].type, val)
                    if not val:
                        groupRow = numericGroupRows.get(str(val), None)
                    else:
                        if not width:
                            binidx = 0
                        elif degenerateBinning:
                            # in degenerate binning, each val has its own bin
                            binidx = numericBins.index((val, val))
                        else:
                            binidx = int((val-minval)//width)
                        groupRow = numericGroupRows[formatRange(numericCols[0], numericBins[min(binidx, nbins-1)])]
                except Exception as e:
                    vd.exceptionCaught(e)

            # add the main bin if no numeric bin (error, or no numeric cols)
            if groupRow is None:
                if numericCols:
                    groupRow = PivotGroupRow(discreteKeys, val, [], {})
                    numericGroupRows[str(val)] = groupRow
                else:
                    groupRow = PivotGroupRow(discreteKeys, (0, 0), [], {})
                    groups[formattedDiscreteKeys] = (numericGroupRows, groupRow)
                self.addRow(groupRow)

            # add the sourcerow to its all bin
            groupRow.sourcerows.append(sourcerow)

            # separate by pivot value
            for col in self.pivotCols:
                varval = col.getTypedValue(sourcerow)
                matchingRows = groupRow.pivotrows.get(varval)
                if matchingRows is None:
                    matchingRows = groupRow.pivotrows[varval] = []
                matchingRows.append(sourcerow)

            if rowfunc:
                rowfunc(groupRow)

    def afterLoad(self):
        super().afterLoad()

        # automatically add cache to all columns now that everything is binned
        for c in self.nonKeyVisibleCols:
            if isinstance(c, AggrColumn):
                c.setCache(True)


@PivotSheet.api
def addcol_aggr(sheet, col):
    hasattr(col, 'origCol') or vd.fail('not an aggregation column')
    for agg_choice in vd.chooseAggregators():
        agg_or_list = vd.aggregators[agg_choice]
        aggs = agg_or_list if isinstance(agg_or_list, list) else [agg_or_list]
        for agg in aggs:
            sheet.addColumnAtCursor(makeAggrColumn(col.origCol, vd.aggregators[agg]))


Sheet.addCommand('W', 'pivot', 'vd.push(makePivot(sheet, keyCols, [cursorCol]))', 'open Pivot Table: group rows by key column and summarize current column')

PivotSheet.addCommand('', 'addcol-aggr', 'addcol_aggr(cursorCol)', 'add aggregation column from source of current column')

vd.addGlobals(
    makePivot=makePivot,
    PivotSheet=PivotSheet,
    PivotGroupRow=PivotGroupRow,
)

vd.addMenuItems('''
    Column > Add column > aggregator > addcol-aggr
    Data > Pivot > pivot
''')



import time

from visidata import vd, VisiData, Progress

from ._ibis import IbisTableSheet, IbisTableIndexSheet, IbisConnectionPool


@VisiData.api
def openurl_clickhouse(vd, p, filetype=None):
    vd.configure_ibis()
    return IbisTableIndexSheet(p.base_stem, source=p, filetype=None, database_name=None,
                               ibis_conpool=IbisConnectionPool(p), sheet_type=ClickhouseSheet)

vd.openurl_clickhouses = vd.openurl_clickhouse


class ClickhouseSheet(IbisTableSheet):
    @property
    def countRows(self):
        if self.total_rows is not None:
            return self.total_rows
        return super().countRows

    def iterload(self):
        with self.con as con:
            qid = None
            try:
                if self.query is None:
                    self.query = self.baseQuery(con)

                self.reloadColumns(self.query, start=0)  # columns based on query without metadata
                sqlstr = con.compile(self.query.limit(self.options.ibis_limit or None))

                with Progress(gerund='clickhousing', sheet=self) as prog:
                    settings = {'max_block_size': 10000}
                    with con.con.query_rows_stream(sqlstr, settings) as stream:
                        prog.total = int(stream.source.summary['total_rows_to_read'])
                        prog.made = 0
                        for row in stream:
                            prog.made += 1
                            yield row
                        self.total_rows = prog.total

            except Exception as e:
                raise
            except BaseException:
                if qid:
                    con._client.cancel(qid)


ClickhouseSheet.init('total_rows', lambda: None)

#ClickhouseSheet.class_options.sql_always_count = True



from copy import copy
import re

from visidata import VisiData, vd, Sheet, options, Column, Progress, anytype, ColumnItem, asyncthread, TypedExceptionWrapper, TypedWrapper, IndexSheet, vlen
from visidata.type_date import date

vd.option('sqlite_onconnect', '', 'sqlite statement to execute after opening a connection')


def requery(url, **kwargs):
    'Return *url* with added or replaced query parameters from *kwargs*.'
    from urllib.parse import urlencode, urlparse, parse_qsl, urlunparse
    url_parts = list(urlparse(url))
    query = dict(parse_qsl(url_parts[4]))
    query.update(kwargs)
    url_parts[4] = urlencode(query)
    return urlunparse(url_parts)


@VisiData.api
def guess_sqlite(vd, p):
    if p.open_bytes().read(16).startswith(b'SQLite format'):
        return dict(filetype='sqlite', _likelihood=10)


@VisiData.api
def open_sqlite(vd, p):
    if not p.is_local():
        vd.fail('sqlite requires an uncompressed, local file')
    return SqliteIndexSheet(p.base_stem, source=p)

@VisiData.api
def openurl_sqlite(vd, p, filetype=None):
    return SqliteIndexSheet(p.base_stem, source=p)

VisiData.open_sqlite3 = VisiData.open_sqlite
VisiData.open_db = VisiData.open_sqlite

# rowdef: list of values
class SqliteSheet(Sheet):
    'Provide functionality for importing SQLite databases.'
    savesToSource = True
    defer = True
    query = ''
    tableName = ''

    def conn(self):
        import sqlite3
        localpath = self.rootSheet().source

        url = localpath if localpath.is_url() else f'file:{localpath.resolve()}'
        url = requery(url, **self.options.getall('sqlite_param_'))

        con = sqlite3.connect(url, uri=True, **self.options.getall('sqlite_connect_'))
        con.text_factory = lambda s, enc=self.options.encoding, encerrs=self.options.encoding_errors: s.decode(enc, encerrs)
        if self.options.sqlite_onconnect:
            con.execute(self.options.sqlite_onconnect)
        return con

    def rawSql(self, q:str) -> 'SqliteSheet':
        return SqliteSheet('query', source=self.source, query=q)

    @property
    def sidebar(self):
        if self.query:
            return '# SQL\n' + self.query
        else:
            return super().sidebar

    def execute(self, conn, sql, parms=None):
        parms = parms or []
        vd.debug(sql)
        return conn.execute(sql, parms)

    def iterload_table(self, tblname:str):
        '''Generate all rows from `tblname` in database at self.source,
        including type information from table_xinfo(), and getting each rowid
        if available (for simpler updates).'''

        def parse_sqlite_type(t):
            m = re.match(r'(\w+)(\((\d+)(,(\d+))?\))?', t.upper())
            if not m: return anytype
            typename, _, i, _, f = m.groups()
            if typename == 'DATE': return date
            if 'INT' in typename: return int
            if typename == 'REAL': return float
            if typename == 'NUMBER':
                return int if f == '0' else float
            return anytype

        self.rowidColumn = None
        with self.conn() as conn:
            if not isinstance(self, SqliteIndexSheet):
                self.columns = []
                for r in self.execute(conn, 'PRAGMA TABLE_XINFO("%s")' % tblname):
                    colnum, colname, coltype, nullable, defvalue, colkey, *_ = r
                    c = ColumnItem(colname, colnum+1, type=parse_sqlite_type(coltype))
                    self.addColumn(c)

                    if colkey:
                        self.setKeys([c])

                sql = self.row[5]  # SQL used to create table
                if 'WITHOUT ROWID' not in sql and 'CREATE VIEW' not in sql:
                    self.rowidColumn = ColumnItem('rowid', 0, type=int, width=0)
                    self.addColumn(self.rowidColumn, index=0)

            if self.rowidColumn:
                r = self.execute(conn, 'SELECT rowid, * FROM "%s"' % tblname)
            else:
                r = self.execute(conn, 'SELECT NULL, * FROM "%s"' % tblname)
            yield from Progress(r, total=r.rowcount-1)

    def iterload_query(self, query:str):
        '''Generate rows from `query` to database at self.source,
        including type information from table_xinfo(), and getting each rowid
        if available (for simpler updates).'''

        with self.conn() as conn:
            self.columns = []
            for c in type(self).columns:
                self.addColumn(copy(c))

            self.result = self.execute(conn, query, parms=getattr(self, 'parms', []))

            for i, desc in enumerate(self.result.description):
                self.addColumn(ColumnItem(desc[0], i))

            for row in self.result:
                yield row

    def iterload(self):
        if self.tableName:
            yield from self.iterload_table(self.tableName)
        elif self.query:
            yield from self.iterload_query(self.query)
        else:
            vd.fail('no query or tablename to load')

    @asyncthread
    def putChanges(self):
        adds, mods, dels = self.getDeferredChanges()
        options_safe_error = options.safe_error
        def value(row, col):
            v = col.getTypedValue(row)
            if isinstance(v, TypedWrapper):
                if isinstance(v, TypedExceptionWrapper):
                    return options_safe_error
                else:
                    return None
            elif not isinstance(v, (int, float, str)):
                v = col.getDisplayValue(r)
            return v

        def values(row, cols):
            vals = []
            for c in cols:
                vals.append(value(row, c))
            return vals

        with self.conn() as conn:
            for r in adds.values():
                cols = self.visibleCols
                sql = 'INSERT INTO "%s" ' % self.tableName
                sql += '(%s)' % ','.join(c.name for c in cols)
                sql += ' VALUES (%s)' % ','.join('?' for c in cols)
                res = self.execute(conn, sql, parms=values(r, cols))
                if res.rowcount != res.arraysize:
                    vd.warning('not all rows inserted') # f'{res.rowcount}/{res.arraysize} rows inserted'

            for row, rowmods in mods.values():
                if not self.rowidColumn:
                    vd.warning('cannot modify rows in tables without rowid')
                    break
                wherecols = [self.rowidColumn]
                sql = 'UPDATE "%s" SET ' % self.tableName
                sql += ', '.join('%s=?' % c.name for c, _ in rowmods.items())
                sql += ' WHERE %s' % ' AND '.join('"%s"=?' % c.name for c in wherecols)
                newvals=values(row, [c for c, _ in rowmods.items()])
                # calcValue gets the 'previous' value (before update)
                wherevals=list(Column.calcValue(c, row) or '' for c in wherecols)
                res = self.execute(conn, sql, parms=newvals+wherevals)
                if res.rowcount != res.arraysize:
                    vd.warning('not all rows updated') # f'{res.rowcount}/{res.arraysize} rows updated'

            for row in dels.values():
                if not self.rowidColumn:
                    vd.warning('cannot delete rows in tables without rowid')
                    break

                wherecols = [self.rowidColumn]
                sql = 'DELETE FROM "%s" ' % self.tableName
                sql += ' WHERE %s' % ' AND '.join('"%s"=?' % c.name for c in wherecols)
                wherevals=list(Column.calcValue(c, row) for c in wherecols)
                res = self.execute(conn, sql, parms=wherevals)
                if res.rowcount != res.arraysize:
                    vd.warning('not all rows deleted') # f'{res.rowcount}/{res.arraysize} rows deleted'

            conn.commit()

        self.preloadHook()
        self.reload()


class SqliteIndexSheet(SqliteSheet, IndexSheet):
    rowtype = 'tables'
    tableName = 'sqlite_master'
    savesToSource = True
    defer = True
    def iterload(self):
        self.addColumn(Column('sql', width=0, getter=lambda c,r:r.row[5]))
        for row in SqliteSheet.iterload(self):
            if row[1] != 'index':
                tblname = row[2]
                yield SqliteSheet(tblname, source=self, tableName=tblname, row=row)

    def putChanges(self):
        adds, mods, dels = self.getDeferredChanges()
        with self.conn() as conn:
            for r in adds.values():
                vd.warning('create a new table by saving a new sheet to this database file')

            for row, rowmods in mods.values():
                cname = self.column('name')
                if len(rowmods) == 1 and cname in rowmods:
                    sql='ALTER TABLE "%s" RENAME TO "%s"' % (cname.calcValue(row), rowmods[cname])
                    self.execute(conn, sql)
                else:
                    vd.warning('can only modify table name')

            for row in dels.values():
                sql = 'DROP TABLE "%s"' % row.tableName
                self.execute(conn, sql)

            conn.commit()

        self.preloadHook()
        self.reload()



@VisiData.api
def save_sqlite(vd, p, *vsheets):
    import sqlite3
    import json
    jsonenc = json.JSONEncoder()  #1589: list/dict values as json

    conn = sqlite3.connect(str(p))
    conn.text_factory = lambda s, enc=vsheets[0].options.encoding: s.decode(enc)
    conn.row_factory = sqlite3.Row
    c = conn.cursor()

    sqltypes = {
        int: 'INTEGER',
        vlen: 'INTEGER',
        float: 'REAL',
        date: 'DATE',
    }

    for t in vd.numericTypes:
        if t not in sqltypes:
            sqltypes[t] = 'REAL'

    for vs in vsheets:
        vs.ensureLoaded()
    vd.sync()

    for vs in vsheets:
        tblname = vd.cleanName(vs.name)
        sqlcols = []
        for col in vs.visibleCols:
            sqlcols.append('"%s" %s' % (col.name, sqltypes.get(col.type, 'TEXT')))
        sql = 'CREATE TABLE IF NOT EXISTS "%s" (%s)' % (tblname, ', '.join(sqlcols))
        c.execute(sql)

        for r in Progress(vs.rows, 'saving'):
            sqlvals = []
            for col in vs.visibleCols:
                v = col.getTypedValue(r)
                if isinstance(v, TypedWrapper):
                    if isinstance(v, TypedExceptionWrapper):
                        v = options.safe_error
                    else:
                        v = None
                elif isinstance(v, (list, tuple, dict)):
                    v = jsonenc.encode(v)
                elif not isinstance(v, (int, float, str)):
                    v = col.getDisplayValue(r)
                sqlvals.append(v)
            sql = 'INSERT INTO "%s" (%s) VALUES (%s)' % (tblname, ','.join(f'"{c.name}"' for c in vs.visibleCols), ','.join('?' for v in sqlvals))
            c.execute(sql, sqlvals)

    conn.commit()


SqliteSheet.addCommand('', 'exec-sql', 'vd.push(rawSql(input("execute SQL: ", type="sql")))', 'execute raw SQL statement')

SqliteIndexSheet.addCommand('a', 'add-table', 'fail("create a new table by saving a sheet to this database file")', 'stub; add table by saving a sheet to the db file instead')
SqliteIndexSheet.bindkey('ga', 'add-table')
SqliteSheet.options.header = 0
VisiData.save_db = VisiData.save_sqlite

vd.addMenuItems('''
    Data > execute SQL query > exec-sql
''')

vd.addGlobals({
    'SqliteIndexSheet': SqliteIndexSheet,
    'SqliteSheet': SqliteSheet,
})



import math
import random

from collections import defaultdict, Counter, OrderedDict
from visidata import vd, asyncthread, ENTER, colors, update_attr, clipdraw, dispwidth
from visidata import BaseSheet, Column, Progress, ColorAttr
from visidata.bezier import bezier

# see www/design/graphics.md

vd.theme_option('disp_graph_labels', True, 'show axes and legend on graph')
vd.theme_option('plot_colors', 'green red yellow cyan magenta white 38 136 168', 'list of distinct colors to use for plotting distinct objects')
vd.theme_option('disp_canvas_charset', ''.join(chr(0x2800+i) for i in range(256)), 'charset to render 2x4 blocks on canvas')
vd.theme_option('disp_graph_pixel_random', False, 'randomly choose attr from set of pixels instead of most common')
vd.theme_option('disp_zoom_incr', 2.0, 'amount to multiply current zoomlevel when zooming')
vd.theme_option('color_graph_hidden', '238 blue', 'color of legend for hidden attribute')
vd.theme_option('color_graph_selected', 'bold', 'color of selected graph points')


class Point:
    def __init__(self, x, y):
        self.x = x
        self.y = y

    def __repr__(self):
        if isinstance(self.x, int):
            return '(%d,%d)' % (self.x, self.y)
        else:
            return '(%.02f,%.02f)' % (self.x, self.y)

    @property
    def xy(self):
        return (self.x, self.y)

class Box:
    def __init__(self, x, y, w=0, h=0):
        self.xmin = x
        self.ymin = y
        self.w = w
        self.h = h

    def __repr__(self):
        return '[%s+%s,%s+%s]' % (self.xmin, self.w, self.ymin, self.h)

    @property
    def xymin(self):
        return Point(self.xmin, self.ymin)

    @property
    def xmax(self):
        return self.xmin + self.w

    @property
    def ymax(self):
        return self.ymin + self.h

    @property
    def center(self):
        return Point(self.xcenter, self.ycenter)

    @property
    def xcenter(self):
        return self.xmin + self.w/2

    @property
    def ycenter(self):
        return self.ymin + self.h/2

    def contains(self, x, y):
        return x >= self.xmin and \
               x < self.xmax and \
               y >= self.ymin and \
               y < self.ymax

def BoundingBox(x1, y1, x2, y2):
    return Box(min(x1, x2), min(y1, y2), abs(x2-x1), abs(y2-y1))


def clipline(x1, y1, x2, y2, xmin, ymin, xmax, ymax):
    'Liang-Barsky algorithm, returns [xn1,yn1,xn2,yn2] of clipped line within given area, or None'
    dx = x2-x1
    dy = y2-y1
    pq = [
        (-dx, x1-xmin),  # left
        ( dx, xmax-x1),  # right
        (-dy, y1-ymin),  # bottom
        ( dy, ymax-y1),  # top
    ]

    u1, u2 = 0, 1
    for p, q in pq:
        if p < 0:  # from outside to inside
            u1 = max(u1, q/p)
        elif p > 0:  # from inside to outside
            u2 = min(u2, q/p)
        else: #  p == 0:  # parallel to bbox
            if q < 0:  # completely outside bbox
                return None

    if u1 > u2:  # completely outside bbox
        return None

    xn1 = x1 + dx*u1
    yn1 = y1 + dy*u1

    xn2 = x1 + dx*u2
    yn2 = y1 + dy*u2

    return xn1, yn1, xn2, yn2

def iterline(x1, y1, x2, y2):
    'Yields (x, y) coords of line from (x1, y1) to (x2, y2)'
    xdiff = abs(x2-x1)
    ydiff = abs(y2-y1)
    xdir = 1 if x1 <= x2 else -1
    ydir = 1 if y1 <= y2 else -1

    r = math.ceil(max(xdiff, ydiff))
    if r == 0:  # point, not line
        yield x1, y1
    else:
        x, y = math.floor(x1), math.floor(y1)
        i = 0
        while i < r:
            x += xdir * xdiff / r
            y += ydir * ydiff / r

            yield x, y
            i += 1


def anySelected(vs, rows):
    for r in rows:
        if vs.isSelected(r):
            return True

#  - width/height are exactly equal to the number of pixels displayable, and can change at any time.
#  - needs to refresh from source on resize
class Plotter(BaseSheet):
    'pixel-addressable display of entire terminal with (x,y) integer pixel coordinates'
    columns=[Column('_')]  # to eliminate errors outside of draw()
    rowtype='pixels'
    def __init__(self, *names, **kwargs):
        super().__init__(*names, **kwargs)
        self.labels = []  # (x, y, text, attr, row)
        self.hiddenAttrs = set()
        self.needsRefresh = False
        self.resetCanvasDimensions(1, 1)  #2171

    @property
    def nRows(self):
        return (self.plotwidth* self.plotheight)

    def resetCanvasDimensions(self, windowHeight, windowWidth):
        'sets total available canvas dimensions to (windowHeight, windowWidth) (in char cells)'
        self.plotwidth = windowWidth*2
        self.plotheight = (windowHeight-1)*4  # exclude status line

        # pixels[y][x] = { attr: list(rows), ... }
        self.pixels = [[defaultdict(list) for x in range(self.plotwidth)] for y in range(self.plotheight)]

    def plotpixel(self, x, y, attr:"str|ColorAttr=''", row=None):
        self.pixels[y][x][attr].append(row)

    def plotline(self, x1, y1, x2, y2, attr:"str|ColorAttr=''", row=None):
        for x, y in iterline(x1, y1, x2, y2):
            self.plotpixel(math.ceil(x), math.ceil(y), attr, row)

    def plotlabel(self, x, y, text, attr:"str|ColorAttr=''", row=None):
        self.labels.append((x, y, text, attr, row))

    def plotlegend(self, i, txt, attr:"str|ColorAttr=''", width=15):
        # move it 1 character to the left b/c the rightmost column can't be drawn to
        self.plotlabel(self.plotwidth-(width+1)*2, i*4, txt, attr)

    @property
    def plotterCursorBox(self):
        'Returns pixel bounds of cursor as a Box.  Override to provide a cursor.'
        return Box(0,0,0,0)

    @property
    def plotterMouse(self):
        return Point(*self.plotterFromTerminalCoord(self.mouseX, self.mouseY))

    def plotterFromTerminalCoord(self, x, y):
        return x*2, y*4

    def getPixelAttrRandom(self, x, y) -> str:
        'weighted-random choice of colornum at this pixel.'
        c = list(attr for attr, rows in self.pixels[y][x].items()
                         for r in rows if attr and attr not in self.hiddenAttrs)
        return random.choice(c) if c else 0

    def getPixelAttrMost(self, x, y) -> str:
        'most common colornum at this pixel.'
        r = self.pixels[y][x]
        if not r:
            return 0
        c = [(len(rows), attr, rows) for attr, rows in r.items() if attr and attr not in self.hiddenAttrs]
        if not c:
            return 0
        _, attr, rows = max(c)
        return attr

    def hideAttr(self, attr:str, hide=True):
        if hide:
            self.hiddenAttrs.add(attr)
        else:
            self.hiddenAttrs.remove(attr)
        self.plotlegends()

    def rowsWithin(self, plotter_bbox, invert_y=False):
        'return list of deduped rows within plotter_bbox'
        ret = {}

        x_start = max(0, plotter_bbox.xmin)
        if len(self.pixels) == 0: return []
        x_end = min(len(self.pixels[0]), plotter_bbox.xmax)

        y_start = max(0, plotter_bbox.ymin)
        y_end = min(len(self.pixels), plotter_bbox.ymax)
        if invert_y:
            y_range = range(y_end-1, y_start-1, -1)
        else:
            y_range = range(y_start, y_end)

        for x in range(x_start, x_end):
            for y in y_range:
                for attr, rows in self.pixels[y][x].items():
                    if attr not in self.hiddenAttrs:
                        for r in rows:
                            ret[self.source.rowid(r)] = r
        return list(ret.values())

    def draw(self, scr):
        windowHeight, windowWidth = scr.getmaxyx()
        if self.needsRefresh:
            self.render(windowHeight, windowWidth)

        self.draw_pixels(scr)
        self.draw_labels(scr)

    def draw_empty(self, scr):
        # use draw_empty() when calling draw_pixels() with clear_empty_squares=False
        cursorBBox = self.plotterCursorBox
        for char_y in range(0, self.plotheight//4):
            for char_x in range(0, self.plotwidth//2):
                cattr = ColorAttr()
                ch = ' '
                # draw cursor
                if cursorBBox.contains(char_x*2, char_y*4) or \
                    cursorBBox.contains(char_x*2+1, char_y*4+3):
                    cattr = update_attr(cattr, colors.color_current_row)
                scr.addstr(char_y, char_x, ch, cattr.attr)

    def draw_pixels(self, scr, clear_empty_squares=True):
        disp_canvas_charset = self.options.disp_canvas_charset or ' o'
        disp_canvas_charset += (256 - len(disp_canvas_charset)) * disp_canvas_charset[-1]
        if self.pixels:
            cursorBBox = self.plotterCursorBox
            getPixelAttr = self.getPixelAttrRandom if self.options.disp_graph_pixel_random else self.getPixelAttrMost

            for char_y in range(0, self.plotheight//4):
                for char_x in range(0, self.plotwidth//2):
                    block_attrs = [
                        getPixelAttr(char_x*2  , char_y*4  ),
                        getPixelAttr(char_x*2  , char_y*4+1),
                        getPixelAttr(char_x*2  , char_y*4+2),
                        getPixelAttr(char_x*2+1, char_y*4  ),
                        getPixelAttr(char_x*2+1, char_y*4+1),
                        getPixelAttr(char_x*2+1, char_y*4+2),
                        getPixelAttr(char_x*2  , char_y*4+3),
                        getPixelAttr(char_x*2+1, char_y*4+3),
                    ]

                    pow2 = 1
                    braille_num = 0
                    for c in block_attrs:
                        if c:
                            braille_num += pow2
                        pow2 *= 2

                    ch = disp_canvas_charset[braille_num]
                    if braille_num != 0:
                        color = Counter(c for c in block_attrs if c).most_common(1)[0][0]
                        cattr = colors.get_color(color)
                    else:
                        cattr = ColorAttr()
                        # don't erase empty squares, useful for subclasses that draw elements like reflines
                        # before pixels are drawn
                        if not clear_empty_squares:
                            continue

                    # draw cursor
                    if cursorBBox.contains(char_x*2, char_y*4) or \
                       cursorBBox.contains(char_x*2+1, char_y*4+3):
                        cattr = update_attr(cattr, colors.color_current_row)

                    if cattr.attr:
                        scr.addstr(char_y, char_x, ch, cattr.attr)

    def draw_labels(self, scr):
        def _mark_overlap_text(labels, textobj):
            def _overlaps(a, b):
                a_x1, _, a_txt, _, _ = a
                b_x1, _, b_txt, _, _ = b
                a_x2 = a_x1 + dispwidth(a_txt)
                b_x2 = b_x1 + dispwidth(b_txt)
                if a_x1 < b_x1 < a_x2 or a_x1 < b_x2 < a_x2 or \
                   b_x1 < a_x1 < b_x2 or b_x1 < a_x2 < b_x2:
                   return True
                else:
                   return False

            label_fldraw = [textobj, True]
            labels.append(label_fldraw)
            for o in labels:
                if _overlaps(o[0], textobj):
                    o[1] = False
                    label_fldraw[1] = False

        if self.options.disp_graph_labels:
            labels_by_line = defaultdict(list) # y -> text labels

            for pix_x, pix_y, txt, attr, row in self.labels:
                if attr in self.hiddenAttrs:
                    continue
                char_y = int(pix_y/4)
                char_x = int(pix_x/2)
                if row is not None:
                    char_x -= math.ceil(dispwidth(txt)/2)*2
                o = (char_x, char_y, txt, attr, row)
                _mark_overlap_text(labels_by_line[char_y], o)

            for line in labels_by_line.values():
                for o, fldraw in line:
                    if fldraw:
                        char_x, char_y, txt, attr, row = o
                        cattr = colors.get_color(attr)
                        clipdraw(scr, char_y, char_x, txt, cattr, dispwidth(txt))
                        cursorBBox = self.plotterCursorBox
                        for c in txt:
                            w = dispwidth(c)
                            # draw cursor if the cursor contains the midpoint of the character cell
                            if cursorBBox.contains(char_x*2+1, char_y*4+2):
                                char_attr = update_attr(cattr, colors.color_current_row)
                                clipdraw(scr, char_y, char_x, c, char_attr, w)
                            char_x += w


# - has a cursor, of arbitrary position and width/height (not restricted to current zoom)
class Canvas(Plotter):
    'zoomable/scrollable virtual canvas with (x,y) coordinates in arbitrary units'
    rowtype = 'plots'
    leftMarginPixels = 10*2
    rightMarginPixels = 4*2
    topMarginPixels = 0*4
    bottomMarginPixels = 1*4  # reserve bottom line for x axis
    guide = '# Canvas\n'

    def __init__(self, *names, **kwargs):
        self.left_margin = self.leftMarginPixels
        super().__init__(*names, **kwargs)

        self.canvasBox = None   # bounding box of entire canvas, in canvas units
        self.visibleBox = None  # bounding box of visible canvas, in canvas units
        self.cursorBox = None   # bounding box of cursor, in canvas units

        self.aspectRatio = 0.0
        self.xzoomlevel = 1.0
        self.yzoomlevel = 1.0
        self.needsRefresh = False

        self.polylines = []   # list of ([(canvas_x, canvas_y), ...], fgcolornum, row)
        self.gridlabels = []  # list of (grid_x, grid_y, label, fgcolornum, row)

        self.legends = OrderedDict()   # txt: attr  (visible legends only)
        self.plotAttrs = {}   # key: attr  (all keys, for speed)
        self.reset()

    @property
    def nRows(self):
        return len(self.polylines)

    def reset(self):
        'clear everything in preparation for a fresh reload()'
        self.polylines.clear()
        self.canvasBox = None
        self.visibleBox = None
        self.cursorBox = None
        self.left_margin = self.leftMarginPixels
        self.legends.clear()
        self.legendwidth = 0
        self.plotAttrs.clear()
        self.unusedAttrs = list(self.options.plot_colors.split())

    def plotColor(self, k) -> str:
        attr = self.plotAttrs.get(k, None)
        if attr is None:
            if self.unusedAttrs:
                attr = self.unusedAttrs.pop(0)
                legend = ' '.join(str(x) for x in k)
            else:
                lastlegend, attr = list(self.legends.items())[-1]
                del self.legends[lastlegend]
                legend = '[other]'

            self.legendwidth = max(self.legendwidth, dispwidth(legend))
            self.legends[legend] = attr
            self.plotAttrs[k] = attr
        return attr

    def resetCanvasDimensions(self, windowHeight, windowWidth):
        old_plotsize = None
        realign_cursor = False
        if hasattr(self, 'plotwidth') and hasattr(self, 'plotheight'):
            old_plotsize = [self.plotheight, self.plotwidth]
            if hasattr(self, 'cursorBox') and self.cursorBox and self.visibleBox:
                # if the cursor is at the origin, realign it with the origin after the resize
                if self.cursorBox.xmin == self.visibleBox.xmin and self.cursorBox.ymin == self.calcBottomCursorY():
                    realign_cursor = True
        super().resetCanvasDimensions(windowHeight, windowWidth)
        # if window is not big enough to contain a particular margin, pretend that margin is 0
        pvbox_x = pvbox_y = 0
        if self.plotwidth > self.left_margin:
            pvbox_x = self.left_margin
        if self.plotheight > self.topMarginPixels:
            pvbox_y = self.topMarginPixels
        if hasattr(self, 'legendwidth'):
            # +4 = 1 empty space after the graph + 2 characters for the legend prefixes of "1:", "2:", etc +
            #      1 character for the empty rightmost column
            new_margin = max(self.rightMarginPixels, (self.legendwidth+4)*2)
            pvbox_xmax = self.plotwidth-new_margin-1
            # ensure the graph data takes up at least 3/4 of the width of the screen no matter how wide the legend gets
            pvbox_xmax = max(pvbox_xmax, math.ceil(self.plotwidth * 3/4)//2*2 + 1)
        else:
            pvbox_xmax = self.plotwidth-self.rightMarginPixels-1
        self.left_margin = min(self.left_margin, math.ceil(self.plotwidth * 1/3)//2*2)
        # enforce a minimum plotview box size of 1x1
        pvbox_xmax = max(pvbox_xmax, 1)
        pvbox_ymax = max(self.plotheight-self.bottomMarginPixels-1, 1)
        self.plotviewBox = BoundingBox(pvbox_x, pvbox_y, pvbox_xmax, pvbox_ymax)
        if [self.plotheight, self.plotwidth] != old_plotsize:
            if hasattr(self, 'cursorBox') and self.cursorBox:
                self.setCursorSizeInPlotterPixels(2, 4)
            if realign_cursor:
                self.cursorBox.ymin = self.calcBottomCursorY()

    @property
    def statusLine(self):
        return 'canvas %s visible %s cursor %s' % (self.canvasBox, self.visibleBox, self.cursorBox)

    @property
    def canvasMouse(self):
        x = self.plotterMouse.x
        y = self.plotterMouse.y
        if not self.canvasBox: return None
        p = Point(self.unscaleX(x), self.unscaleY(y))
        return p

    def setCursorSize(self, p):
        'sets width based on diagonal corner p'
        if not p: return
        self.cursorBox = BoundingBox(self.cursorBox.xmin, self.cursorBox.ymin, p.x, p.y)
        self.cursorBox.w = max(self.cursorBox.w, self.canvasCharWidth)
        self.cursorBox.h = max(self.cursorBox.h, self.canvasCharHeight)

    def setCursorSizeInPlotterPixels(self, w, h):
        self.setCursorSize(Point(self.cursorBox.xmin + w/2 * self.canvasCharWidth,
                                 self.cursorBox.ymin + h/4 * self.canvasCharHeight))

    def formatX(self, v):
        return str(v)

    def formatY(self, v):
        return str(v)

    def commandCursor(sheet, execstr):
        'Return (col, row) of cursor suitable for cmdlog replay of execstr.'
        contains = lambda s, *substrs: any((a in s) for a in substrs)
        colname, rowname = '', ''
        if contains(execstr, 'plotterCursorBox'):
            bb = sheet.cursorBox
            colname = '%s %s' % (sheet.formatX(bb.xmin), sheet.formatX(bb.xmax))
            rowname = '%s %s' % (sheet.formatY(bb.ymin), sheet.formatY(bb.ymax))
        elif contains(execstr, 'plotterVisibleBox'):
            bb = sheet.visibleBox
            colname = '%s %s' % (sheet.formatX(bb.xmin), sheet.formatX(bb.xmax))
            rowname = '%s %s' % (sheet.formatY(bb.ymin), sheet.formatY(bb.ymax))
        return colname, rowname

    @property
    def canvasCharWidth(self):
        'Width in canvas units of a single char in the terminal'
        return self.visibleBox.w*2/self.plotviewBox.w

    @property
    def canvasCharHeight(self):
        'Height in canvas units of a single char in the terminal'
        return self.visibleBox.h*4/self.plotviewBox.h

    @property
    def plotterVisibleBox(self):
        return BoundingBox(self.scaleX(self.visibleBox.xmin),
                           self.scaleY(self.visibleBox.ymin),
                           self.scaleX(self.visibleBox.xmax),
                           self.scaleY(self.visibleBox.ymax))

    @property
    def plotterCursorBox(self):
        if self.cursorBox is None:
            return Box(0,0,0,0)
        return BoundingBox(self.scaleX(self.cursorBox.xmin),
                           self.scaleY(self.cursorBox.ymin),
                           self.scaleX(self.cursorBox.xmax),
                           self.scaleY(self.cursorBox.ymax))

    def startCursor(self):
        cm = self.canvasMouse
        if cm:
            self.cursorBox = Box(*cm.xy)
            return True
        else:
            return None

    def point(self, x, y, attr:"str|ColorAttr=''", row=None):
        self.polylines.append(([(x, y)], attr, row))

    def line(self, x1, y1, x2, y2, attr:"str|ColorAttr=''", row=None):
        self.polylines.append(([(x1, y1), (x2, y2)], attr, row))

    def polyline(self, vertexes, attr:"str|ColorAttr=''", row=None):
        'adds lines for (x,y) vertexes of a polygon'
        self.polylines.append((vertexes, attr, row))

    def polygon(self, vertexes, attr:"str|ColorAttr=''", row=None):
        'adds lines for (x,y) vertexes of a polygon'
        self.polylines.append((vertexes + [vertexes[0]], attr, row))

    def qcurve(self, vertexes, attr:"str|ColorAttr=''", row=None):
        'Draw quadratic curve from vertexes[0] to vertexes[2] with control point at vertexes[1]'
        if len(vertexes) != 3:
            vd.fail('need exactly 3 points for qcurve (got %d)' % len(vertexes))

        x1, y1 = vertexes[0]
        x2, y2 = vertexes[1]
        x3, y3 = vertexes[2]

        for x, y in bezier(x1, y1, x2, y2, x3, y3):
            self.point(x, y, attr, row)

    def label(self, x, y, text, attr:"str|ColorAttr=''", row=None):
        self.gridlabels.append((x, y, text, attr, row))

    def fixPoint(self, plotterPoint, canvasPoint):
        'adjust visibleBox.xymin so that canvasPoint is plotted at plotterPoint'
        self.visibleBox.xmin = canvasPoint.x - self.canvasW(plotterPoint.x-self.plotviewBox.xmin)
        self.visibleBox.ymin = canvasPoint.y - self.canvasH(plotterPoint.y-self.plotviewBox.ymin)
        self.resetBounds()

    def zoomTo(self, bbox):
        'set visible area to bbox, maintaining aspectRatio if applicable'
        self.fixPoint(self.plotviewBox.xymin, bbox.xymin)
        self.xzoomlevel=bbox.w/self.canvasBox.w
        self.yzoomlevel=bbox.h/self.canvasBox.h
        self.resetBounds()

    def incrZoom(self, incr):
        self.xzoomlevel *= incr
        self.yzoomlevel *= incr

        self.resetBounds()

    def resetBounds(self, refresh=True):
        'create canvasBox and cursorBox if necessary, and set visibleBox w/h according to zoomlevels.  then redisplay legends.'
        if not self.canvasBox:
            xmin, ymin, xmax, ymax = None, None, None, None
            for vertexes, attr, row in self.polylines:
                for x, y in vertexes:
                    if xmin is None or x < xmin: xmin = x
                    if ymin is None or y < ymin: ymin = y
                    if xmax is None or x > xmax: xmax = x
                    if ymax is None or y > ymax: ymax = y
            xmin = xmin or 0
            xmax = xmax or 0
            ymin = ymin or 0
            ymax = ymax or 0
            if xmin == xmax:
                xmax += 1
            if ymin == ymax:
                ymax += 1
            self.canvasBox = BoundingBox(float(xmin), float(ymin), float(xmax), float(ymax))

        w = self.calcVisibleBoxWidth()
        h = self.calcVisibleBoxHeight()
        if not self.visibleBox:
            # initialize minx/miny, but w/h must be set first to center properly
            self.visibleBox = Box(0, 0, w, h)
            self.visibleBox.xmin = self.canvasBox.xmin + (self.canvasBox.w / 2) * (1 - self.xzoomlevel)
            self.visibleBox.ymin = self.canvasBox.ymin + (self.canvasBox.h / 2) * (1 - self.yzoomlevel)
        else:
            self.visibleBox.w = w
            self.visibleBox.h = h

        if not self.cursorBox:
            cb_xmin = self.visibleBox.xmin
            cb_ymin = self.calcBottomCursorY()
            self.cursorBox = Box(cb_xmin, cb_ymin, self.canvasCharWidth, self.canvasCharHeight)

        self.plotlegends()
        if refresh:
            self.refresh()

    def calcTopCursorY(self):
        'ymin for the cursor that will align its top with the top edge of the graph'
        # + (1/4*self.canvasCharHeight) shifts the cursor up by 1 plotter pixel.
        # That shift makes the cursor contain the top data point.
        # Otherwise, the top data point would have y == plotterCursorBox.ymax,
        # which would not be inside plotterCursorBox. Shifting the cursor makes
        # plotterCursorBox.ymax > y for that top point.
        return self.visibleBox.ymax - self.cursorBox.h + (1/4*self.canvasCharHeight)

    def calcBottomCursorY(self):
        'ymin for the cursor that will align its bottom with the bottom edge of the graph'
        return self.visibleBox.ymin

    def plotlegends(self):
        # display labels
        for i, (legend, attr) in enumerate(self.legends.items()):
            self.addCommand(str(i+1), f'toggle-{i+1}', f'hideAttr("{attr}", "{attr}" not in hiddenAttrs)', f'toggle display of "{legend}"')
            if attr in self.hiddenAttrs:
                attr = 'graph_hidden'
            # add 2 characters to width to account for '1:' '2:' etc
            self.plotlegend(i, '%s:%s'%(i+1,legend), attr, width=self.legendwidth+2)

    def checkCursor(self):
        'override Sheet.checkCursor'
        if self.visibleBox and self.cursorBox:
            if self.cursorBox.h < self.canvasCharHeight:
                self.cursorBox.h = self.canvasCharHeight*3/4
            if self.cursorBox.w < self.canvasCharWidth:
                self.cursorBox.w = self.canvasCharWidth*3/4

        return False

    @property
    def xScaler(self):
        xratio = self.plotviewBox.w/(self.canvasBox.w*self.xzoomlevel)
        if self.aspectRatio:
            yratio = self.plotviewBox.h/(self.canvasBox.h*self.yzoomlevel)
            return self.aspectRatio*min(xratio, yratio)
        else:
            return xratio

    @property
    def yScaler(self):
        yratio = self.plotviewBox.h/(self.canvasBox.h*self.yzoomlevel)
        if self.aspectRatio:
            xratio = self.plotviewBox.w/(self.canvasBox.w*self.xzoomlevel)
            return min(xratio, yratio)
        else:
            return yratio

    def calcVisibleBoxWidth(self):
        w = self.canvasBox.w * self.xzoomlevel
        if self.aspectRatio:
            h = self.canvasBox.h * self.yzoomlevel
            xratio = self.plotviewBox.w / w
            yratio = self.plotviewBox.h / h
            if xratio <= yratio:
                return w / self.aspectRatio
            else:
                return self.plotviewBox.w / (self.aspectRatio * yratio)
        else:
            return w

    def calcVisibleBoxHeight(self):
        h = self.canvasBox.h * self.yzoomlevel
        if self.aspectRatio:
            w = self.canvasBox.w * self.yzoomlevel
            xratio = self.plotviewBox.w / w
            yratio = self.plotviewBox.h / h
            if xratio < yratio:
                return self.plotviewBox.h / xratio
            else:
                return h
        else:
            return h

    def scaleX(self, canvasX) -> int:
        'returns a plotter x coordinate'
        return self.plotviewBox.xmin+round((canvasX-self.visibleBox.xmin)*self.xScaler)

    def scaleY(self, canvasY) -> int:
        'returns a plotter y coordinate'
        return self.plotviewBox.ymin+round((canvasY-self.visibleBox.ymin)*self.yScaler)

    def unscaleX(self, plotterX):
        'performs the inverse of scaleX, returns a canvas x coordinate'
        return (plotterX-self.plotviewBox.xmin)/self.xScaler + self.visibleBox.xmin

    def unscaleY(self, plotterY):
        'performs the inverse of scaleY, returns a canvas y coordinate'
        return (plotterY-self.plotviewBox.ymin)/self.yScaler + self.visibleBox.ymin

    def canvasW(self, plotter_width):
        'plotter X units to canvas units'
        return plotter_width/self.xScaler

    def canvasH(self, plotter_height):
        'plotter Y units to canvas units'
        return plotter_height/self.yScaler

    def refresh(self):
        'triggers render() on next draw()'
        self.needsRefresh = True

    def render(self, h, w):
        'resets plotter, cancels previous render threads, spawns a new render'
        self.needsRefresh = False
        vd.cancelThread(*(t for t in self.currentThreads if t.name == 'plotAll_async'))
        self.labels.clear()
        self.resetCanvasDimensions(h, w)
        self.resetBounds(refresh=False)
        self.render_async()

    @asyncthread
    def render_async(self):
        self.plot_elements()

    def plot_elements(self, invert_y=False):
        'plots points and lines and text onto the plotter'

        self.resetBounds(refresh=False)

        bb = self.visibleBox
        xmin, ymin, xmax, ymax = bb.xmin, bb.ymin, bb.xmax, bb.ymax
        xfactor, yfactor = self.xScaler, self.yScaler
        plotxmin = self.plotviewBox.xmin
        if invert_y:
            plotymax = self.plotviewBox.ymax
        else:
            plotymin = self.plotviewBox.ymin

        for vertexes, attr, row in Progress(self.polylines, 'rendering'):
            if len(vertexes) == 1:  # single point
                x1, y1 = vertexes[0]
                x1, y1 = float(x1), float(y1)
                if xmin <= x1 <= xmax and ymin <= y1 <= ymax:
                    # equivalent to self.scaleX(x1) and self.scaleY(y1), inlined for speed
                    x = plotxmin+round((x1-xmin)*xfactor)
                    if invert_y:
                        y = plotymax-round((y1-ymin)*yfactor)
                    else:
                        y = plotymin+round((y1-ymin)*yfactor)
                    self.plotpixel(x, y, attr, row)
                continue

            prev_x, prev_y = vertexes[0]
            for x, y in vertexes[1:]:
                r = clipline(prev_x, prev_y, x, y, xmin, ymin, xmax, ymax)
                if r:
                    x1, y1, x2, y2 = r
                    x1 = plotxmin+float(x1-xmin)*xfactor
                    x2 = plotxmin+float(x2-xmin)*xfactor
                    if invert_y:
                        y1 = plotymax-float(y1-ymin)*yfactor
                        y2 = plotymax-float(y2-ymin)*yfactor
                    else:
                        y1 = plotymin+float(y1-ymin)*yfactor
                        y2 = plotymin+float(y2-ymin)*yfactor
                    self.plotline(x1, y1, x2, y2, attr, row)
                prev_x, prev_y = x, y

        for x, y, text, attr, row in Progress(self.gridlabels, 'labeling'):
            self.plotlabel(self.scaleX(x), self.scaleY(y), text, attr, row)

    @asyncthread
    def deleteSourceRows(self, rows):
        rows = list(rows)
        self.source.copyRows(rows)
        self.source.deleteBy(lambda r,rows=rows: r in rows)
        self.reload()

Plotter.addCommand('v', 'visibility', 'options.disp_graph_labels = not options.disp_graph_labels', 'toggle disp_graph_labels option')

Canvas.addCommand(None, 'go-left', 'if cursorBox: sheet.cursorBox.xmin -= cursorBox.w', 'move cursor left by its width')
Canvas.addCommand(None, 'go-right', 'if cursorBox: sheet.cursorBox.xmin += cursorBox.w', 'move cursor right by its width' )
Canvas.addCommand(None, 'go-up', 'if cursorBox: sheet.cursorBox.ymin -= cursorBox.h', 'move cursor up by its height')
Canvas.addCommand(None, 'go-down', 'if cursorBox: sheet.cursorBox.ymin += cursorBox.h', 'move cursor down by its height')
Canvas.addCommand(None, 'go-leftmost', 'if cursorBox: sheet.cursorBox.xmin = visibleBox.xmin', 'move cursor to left edge of visible canvas')
Canvas.addCommand(None, 'go-rightmost', 'if cursorBox: sheet.cursorBox.xmin = visibleBox.xmax-cursorBox.w+(1/2*canvasCharWidth)', 'move cursor to right edge of visible canvas')
Canvas.addCommand(None, 'go-top',    'if cursorBox: sheet.cursorBox.ymin = sheet.calcTopCursorY()', 'move cursor to top edge of visible canvas')
Canvas.addCommand(None, 'go-bottom', 'if cursorBox: sheet.cursorBox.ymin = sheet.calcBottomCursorY()', 'move cursor to bottom edge of visible canvas')

Canvas.addCommand(None, 'go-pagedown', 't=(visibleBox.ymax-visibleBox.ymin); sheet.cursorBox.ymin += t; sheet.visibleBox.ymin += t; refresh()', 'move cursor down to next visible page')
Canvas.addCommand(None, 'go-pageup', 't=(visibleBox.ymax-visibleBox.ymin); sheet.cursorBox.ymin -= t; sheet.visibleBox.ymin -= t; refresh()', 'move cursor up to previous visible page')

Canvas.addCommand('zh', 'go-left-small', 'sheet.cursorBox.xmin -= canvasCharWidth', 'move cursor left one character')
Canvas.addCommand('zl', 'go-right-small', 'sheet.cursorBox.xmin += canvasCharWidth', 'move cursor right one character')
Canvas.addCommand('zj', 'go-down-small', 'sheet.cursorBox.ymin += canvasCharHeight', 'move cursor down one character')
Canvas.addCommand('zk', 'go-up-small', 'sheet.cursorBox.ymin -= canvasCharHeight', 'move cursor up one character')

Canvas.addCommand('gH', 'resize-cursor-halfwide', 'sheet.cursorBox.w /= 2', 'halve cursor width')
Canvas.addCommand('gL', 'resize-cursor-doublewide', 'sheet.cursorBox.w *= 2', 'double cursor width')
Canvas.addCommand('gJ','resize-cursor-halfheight', 'sheet.cursorBox.h /= 2', 'halve cursor height')
Canvas.addCommand('gK', 'resize-cursor-doubleheight', 'sheet.cursorBox.h *= 2', 'double cursor height')

Canvas.addCommand('H', 'resize-cursor-thinner', 'sheet.cursorBox.w -= canvasCharWidth', 'decrease cursor width by one character')
Canvas.addCommand('L', 'resize-cursor-wider', 'sheet.cursorBox.w += canvasCharWidth', 'increase cursor width by one character')
Canvas.addCommand('J', 'resize-cursor-taller', 'sheet.cursorBox.h += canvasCharHeight', 'increase cursor height by one character')
Canvas.addCommand('K', 'resize-cursor-shorter', 'sheet.cursorBox.h -= canvasCharHeight', 'decrease cursor height by one character')
Canvas.addCommand('zz', 'zoom-cursor', 'zoomTo(cursorBox)', 'set visible bounds to cursor')

Canvas.addCommand('-', 'zoomout-cursor', 'tmp=cursorBox.center; incrZoom(options.disp_zoom_incr); fixPoint(plotviewBox.center, tmp)', 'zoom out from cursor center')
Canvas.addCommand('+', 'zoomin-cursor', 'tmp=cursorBox.center; incrZoom(1.0/options.disp_zoom_incr); fixPoint(plotviewBox.center, tmp)', 'zoom into cursor center')
Canvas.addCommand('_', 'zoom-all', 'sheet.canvasBox = None; sheet.visibleBox = None; sheet.xzoomlevel=sheet.yzoomlevel=1.0; resetBounds()', 'zoom to fit full extent')
Canvas.addCommand('z_', 'set-aspect', 'sheet.aspectRatio = float(input("aspect ratio=", value=aspectRatio)); refresh()', 'set aspect ratio')

# set cursor box with left click
Canvas.addCommand('BUTTON1_PRESSED', 'start-cursor', 'startCursor()', 'start cursor box with left mouse button press')
Canvas.addCommand('BUTTON1_RELEASED', 'end-cursor', 'cm=canvasMouse; setCursorSize(cm) if cm else None', 'end cursor box with left mouse button release')
Canvas.addCommand('BUTTON1_CLICKED', 'remake-cursor', 'startCursor(); cm=canvasMouse; setCursorSize(cm) if cm else None', 'end cursor box with left mouse button release')
Canvas.bindkey('BUTTON1_DOUBLE_CLICKED', 'remake-cursor')
Canvas.bindkey('BUTTON1_TRIPLE_CLICKED', 'remake-cursor')

Canvas.addCommand('BUTTON3_PRESSED', 'start-move', 'cm=canvasMouse; sheet.anchorPoint = cm if cm else None', 'mark grid point to move')
Canvas.addCommand('BUTTON3_RELEASED', 'end-move', 'fixPoint(plotterMouse, anchorPoint) if anchorPoint else None', 'mark canvas anchor point')
# A click does not actually move the canvas, but gives useful UI feedback. It helps users understand that they can do press-drag-release.
Canvas.addCommand('BUTTON3_CLICKED', 'move-canvas',  '', 'move canvas (in place)')
Canvas.bindkey('BUTTON3_DOUBLE_CLICKED', 'move-canvas')
Canvas.bindkey('BUTTON3_TRIPLE_CLICKED', 'move-canvas')

Canvas.addCommand('ScrollUp', 'zoomin-mouse', 'cm=canvasMouse; incrZoom(1.0/options.disp_zoom_incr) if cm else fail("cannot zoom in on unplotted canvas"); fixPoint(plotterMouse, cm)', 'zoom in with scroll wheel')
Canvas.addCommand('ScrollDown', 'zoomout-mouse', 'cm=canvasMouse; incrZoom(options.disp_zoom_incr) if cm else fail("cannot zoom out on unplotted canvas"); fixPoint(plotterMouse, cm)', 'zoom out with scroll wheel')

Canvas.addCommand('s', 'select-cursor', 'source.select(list(rowsWithin(plotterCursorBox)))', 'select rows on source sheet contained within canvas cursor')
Canvas.addCommand('t', 'stoggle-cursor', 'source.toggle(list(rowsWithin(plotterCursorBox)))', 'toggle selection of rows on source sheet contained within canvas cursor')
Canvas.addCommand('u', 'unselect-cursor', 'source.unselect(list(rowsWithin(plotterCursorBox)))', 'unselect rows on source sheet contained within canvas cursor')
Canvas.addCommand(ENTER, 'dive-cursor', 'vs=copy(source); vs.rows=list(rowsWithin(plotterCursorBox)); vd.push(vs)', 'open sheet of source rows contained within canvas cursor')
Canvas.addCommand('d', 'delete-cursor', 'deleteSourceRows(rowsWithin(plotterCursorBox))', 'delete rows on source sheet contained within canvas cursor')

Canvas.addCommand('gs', 'select-visible', 'source.select(list(rowsWithin(plotterVisibleBox)))', 'select rows on source sheet visible on screen')
Canvas.addCommand('gt', 'stoggle-visible', 'source.toggle(list(rowsWithin(plotterVisibleBox)))', 'toggle selection of rows on source sheet visible on screen')
Canvas.addCommand('gu', 'unselect-visible', 'source.unselect(list(rowsWithin(plotterVisibleBox)))', 'unselect rows on source sheet visible on screen')
Canvas.addCommand('g'+ENTER, 'dive-visible', 'vs=copy(source); vs.rows=list(rowsWithin(plotterVisibleBox)); vd.push(vs)', 'open sheet of source rows visible on screen')
Canvas.addCommand('gd', 'delete-visible', 'deleteSourceRows(rowsWithin(plotterVisibleBox))', 'delete rows on source sheet visible on screen')

vd.addGlobals({
    'Canvas': Canvas,
    'Plotter': Plotter,
    'BoundingBox': BoundingBox,
    'Box': Box,
    'Point': Point,
})

vd.addMenuItems('''
    Plot > Resize cursor > height > double > resize-cursor-doubleheight
    Plot > Resize cursor > height > half > resize-cursor-halfheight
    Plot > Resize cursor > height > shorter > resize-cursor-shorter
    Plot > Resize cursor > height > taller > resize-cursor-taller
    Plot > Resize cursor > width > double > resize-cursor-doublewide
    Plot > Resize cursor > width > half > resize-cursor-halfwide
    Plot > Resize cursor > width > thinner > resize-cursor-thinner
    Plot > Resize cursor > width > wider > resize-cursor-wider
    Plot > Resize graph > X axis > resize-x-input
    Plot > Resize graph > Y axis > resize-y-input
    Plot > Resize graph > aspect ratio > set-aspect
    Plot > Zoom > out > zoomout-cursor
    Plot > Zoom > in > zoomin-cursor
    Plot > Zoom > cursor > zoom-all
    Plot > Dive into cursor > dive-cursor
    Plot > Delete > under cursor > delete-cursor
''')



'''#1488
Usage (in .visidatarc):


    DirSheet.knownCols.directory.width = 0
    Sheet.knownCols.date.type = date
'''

from visidata import Sheet, DefaultAttrDict


Sheet.knownCols = DefaultAttrDict()


@Sheet.before
def afterLoad(sheet):
    for colname, attrs in sheet.knownCols.items():
        col = sheet.colsByName.get(colname)
        if col:
            for k, v in attrs.items():
                setattr(col, k, v)



__author__ = "Paul McCann <polm@dampfkraft.com>"

from visidata import vd, VisiData, TableSheet, ItemColumn

@VisiData.api
def open_conll(vd, p):
    return ConllSheet(p.base_stem, source=p)


@VisiData.api
def open_conllu(vd, p):
    return ConllSheet(p.base_stem, source=p)


class ConllSheet(TableSheet):
    rowtype='tokens'
    # see here for reference:
    # https://universaldependencies.org/format.html
    columns=[
        # Usually an integer, but can be prefixed like "dev-s1"
        ItemColumn('sent_id', 0, type=str),
        # token ID is almost always an integer, but can be technically be a decimal between 0 and 1.
        # starts from 1 for each sentence.
        ItemColumn('token_id', 1, type=int),
        # form from the raw input, aka surface
        ItemColumn('form', 2, type=str),
        ItemColumn('lemma', 3, type=str),
        ItemColumn('upos', 4, type=str),
        ItemColumn('xpos', 5, type=str),
        ItemColumn('feats', 6, type=dict),
        ItemColumn('head', 7, type=int),
        ItemColumn('deprel', 8, type=str),
        # possibly list of pairs, but often? unused
        ItemColumn('deps', 9),
        # empty or a dictionary
        ItemColumn('misc', 10, type=dict),
    ]
    def iterload(self):
        pyconll = vd.importExternal('pyconll')

        # sent_id + token_id will be unique
        self.setKeys([self.columns[0], self.columns[1]])

        with self.source.open(encoding='utf-8') as fp:
            for sent in pyconll.load.iter_sentences(fp):
                sent_id = sent.id
                for token in sent:
                    yield [sent_id, token.id, token._form, token.lemma, token.upos,
                            token.xpos, token.feats, token.head, token.deprel, token.deps, token.misc]
